/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 * vmx86.c --
 *
 *     Platform independent routines for creating/destroying/running
 *     virtual machine monitors.
 *
 */

#ifdef linux
/* Must come before any kernel header file --hpreg */
#   include "driver-config.h"

#   include <linux/string.h> /* memset() in the kernel */
#   include <linux/sched.h> /* jiffies from the kernel */
#else
#   include <string.h>
#endif

#include "vmware.h"
#include "x86.h"
#include "vmx86.h"
#include "task.h"
#include "initblock.h"
#include "vm_asm.h"
#include "iocontrols.h"
#ifdef USE_PERFCTRS_HOSTED
#   include "perfctr.h"
#endif
#include "hostif.h"
#include "memtrack.h"
#include "hash.h"
#if defined(_WIN64)
#include "vmmon-asm-x86-64.h"
#endif
#include "x86vt.h"
#include "x86svm.h"
#if defined(linux)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0) && \
    (!defined(VM_X86_64) || LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9))
#include <asm/timex.h>
#define VMW_HAS_CPUKHZ
#endif
#endif
#include "modulecall_compat.h"

/*
 * Dynamic structures for keeping NUMA info. 
 */
static unsigned int     numaNumNodes;
static unsigned int     numaNumMemRanges;
static NUMA_NodeInfo    *numaNodes;
static uint32           *apicToNUMANode;
static NUMA_MemRange    *numaMemRangesList;

/*
 * Keep track of the virtual machines that have been
 * created using the following structures.
 *
 */

static VMDriver *vmDriverList = NULL;

static LockedPageLimit lockedPageLimit = {
   0,                // host: does not need to be initialized.
   0,                // configured: must be set by some VM as it is powered on.
   MAX_LOCKED_PAGES, // dynamic
};

/* Percentage of guest "paged" memory that must fit within the hard limit. */
static unsigned minVmMemPct;

/* Number of pages actually locked by all virtual machines */
static unsigned numLockedPages;

/* Total virtual machines on this host */
static unsigned vmCount;

/* Track the state of a shared page with COW page frame */
typedef struct COWPage {
   struct COWPage *next;
   uint64 key;            // Hash of this MPN
   MPN sharedMPN;         // Anon MPN shared between vms
   uint32 ref[MAX_VMS];   // Per-vm ref count 
   Bool locked;           // Keep entry even when refcount -> 0
} COWPage;

/* Track promising candidates for sharing with COW hint frames */
typedef struct COWHint {
   struct COWHint *nextByKey; // Links COW hint frames by hash on page contents
   struct COWHint *prevByKey; // Back pointer
   struct COWHint *nextByMPN; // Links COW hint frames by page number
   struct COWHint *prevByMPN; // Back pointer 
   uint64 key;                // Hash of origMPN
   BPN bpn;                   // Guest BPN
   MPN origMPN;               // Guest MPN
   VMDriver *vm;              // VM containing BPN/MPN pair
} COWHint;

/* Chaining hash tables to store pages and hashes */
#define COW_MAX_HINTS           0x100000
#define COW_HINT_TABLE_KEY_SIZE 0x1000
#define COW_HINT_TABLE_KEY_MASK (COW_HINT_TABLE_KEY_SIZE - 1)
#define COW_HINT_TABLE_MPN_SIZE 0x1000
#define COW_HINT_TABLE_MPN_MASK (COW_HINT_TABLE_MPN_SIZE - 1)

#define COW_PAGE_TABLE_SIZE     0x1000
#define COW_PAGE_TABLE_MASK     (COW_PAGE_TABLE_SIZE - 1)

// COW page table walk flags
typedef enum {
   COWMatchKey,
   COWMatchStale,
   COWMatchLocked
} COWPageWalkArgs;

static COWPage *cowPageTable[COW_PAGE_TABLE_SIZE];
static COWHint *cowHintTableKey[COW_HINT_TABLE_KEY_SIZE];
static COWHint *cowHintTableMPN[COW_HINT_TABLE_MPN_SIZE];

/* COW-sharing statistics */
static struct {
   uint32 numHints;
   uint32 numBreaks;       // number of times COW was broken
   uint32 totalUniqueMPNs; // total MPNs ever used as shared pages
   uint32 uniqueMPNs;      // current count of unique shared pages
} cowState;

/* Frame for preallocated zero COW page */
static COWPage *cowZeroPage;

/* Classes of page contents */
typedef enum {
   PCID_UNKNOWN,
   PCID_ZERO,
   PCID_WORD5,
   PCID_PERIODIC_1,
   PCID_PERIODIC_3,
   PCID_PERIODIC_4
} PageClassID;

/* Signature of a page */
typedef union {
   uint64 key;
   PageClassID id;
} PageSig;

/* Temp arrays to hold cow page content */
static char cowBuf[PAGE_SIZE];
static char cowContents[MAX_VMS][PAGE_SIZE];

/* 
 * Allocation of ref locations insdie the COWPage struct.
 * We implement a list of allocated cowID's using an array.  
 * The array is initialized with the values 1...MAX_VMS-1, INVALID_COWID.
 * cowIDsAllocated holds the last id given out and cowIDsUnused
 * holds the next id to give out.  
 */

#define INVALID_COWID (-1)
static int cowIDList[MAX_VMS];
static int cowIDsAllocated;
static int cowIDsUnused;

/* Max rate requested for fast clock by any virtual machine. */
static unsigned globalFastClockRate;


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageTableChainHeadPtr --
 *
 *       For the hash table that keeps COW page frames returns pointer to
 *       the head of the chain corresponding to the key value specified.
 *
 * Results:
 *       Pointer to chain head.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE COWPage**
Vmx86COWPageTableChainHeadPtr(uint64 key)
{
   unsigned slot = (unsigned) key & COW_PAGE_TABLE_MASK;
   return &cowPageTable[slot];
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintTableKeyChainHeadPtr --
 *
 *       For the hash table that keeps COW hint frames returns pointer to 
 *       the head of the chain corresponding to the key value specified.
 *
 * Results:
 *       Pointer to chain head.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE COWHint**
Vmx86COWHintTableKeyChainHeadPtr(uint64 key)
{
   unsigned slot = (unsigned) key & COW_HINT_TABLE_KEY_MASK;
   return &cowHintTableKey[slot];
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintTableMPNChainHeadPtr --
 *
 *       For the hash table that keeps COW hint frames returns pointer to 
 *       the head of the chain corresponding to the page number specified.
 *
 * Results:
 *       Pointer to chain head.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE COWHint**
Vmx86COWHintTableMPNChainHeadPtr(MPN mpn)
{
   unsigned slot = (unsigned) mpn & COW_HINT_TABLE_MPN_MASK;
   return &cowHintTableMPN[slot];
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageTableAdd --
 *
 *       Links COW page frame to the corresponding hash table. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       New element is added to a COW page hash table.
 *
 *----------------------------------------------------------------------
 */

static INLINE void
Vmx86COWPageTableAdd(COWPage *page)
{
   COWPage **head;
   ASSERT(page != NULL);
   head = Vmx86COWPageTableChainHeadPtr(page->key);
   page->next = *head;
   *head = page;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageTableRemove --
 *
 *       Unlinks COW page frame from the corresponding hash table. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       New element is removed from a COW page hash table.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86COWPageTableRemove(COWPage *page, COWPage *prev)
{
   COWPage *next;

   ASSERT(page != NULL);
   ASSERT(prev == NULL || prev->next == page);

   next = page->next;
   if (prev != NULL) {
      prev->next = next;
   } else {
      *Vmx86COWPageTableChainHeadPtr(page->key) = next;
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintTableKeyAdd --
 *
 *       Links COW hint frame to the hash table that keeps track of page
 *       contents. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       New element is added to the corresponding COW hint hash table.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86COWHintTableKeyAdd(COWHint *hint)
{
   COWHint **head;
   ASSERT(hint != NULL);
   head = Vmx86COWHintTableKeyChainHeadPtr(hint->key);
   hint->nextByKey = *head;
   hint->prevByKey = NULL;
   if (*head != NULL) {
      (*head)->prevByKey = hint;
   }
   *head = hint;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintTableMPNAdd --
 *
 *       Links COW hint frame to the hash table that keeps track of page
 *       numbers.
 *
 * Results:
 *       None
 *
 * Side effects:
 *       New element is added to the corresponding COW hint hash table.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86COWHintTableMPNAdd(COWHint *hint)
{
   COWHint **head;
   ASSERT(hint != NULL);
   head = Vmx86COWHintTableMPNChainHeadPtr(hint->origMPN);
   hint->nextByMPN = *head;
   hint->prevByMPN = NULL;
   if (*head != NULL) {
      (*head)->prevByMPN = hint;
   }
   *head = hint;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintStorageAdd --
 *
 *       Links COW hint frame to both hash tables. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       New element is added to both COW hint hash tables.
 *
 *----------------------------------------------------------------------
 */

static INLINE void
Vmx86COWHintStorageAdd(COWHint *hint)
{
   Vmx86COWHintTableKeyAdd(hint);
   Vmx86COWHintTableMPNAdd(hint);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintTableKeyRemove --
 *
 *       Unlinks COW hint frame from the hash table that keeps track of 
 *       page contents.
 *
 * Results:
 *       None
 *
 * Side effects:
 *       Element specified is removed from the corresponding COW hint hash 
 *       table.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86COWHintTableKeyRemove(COWHint *hint)
{
   COWHint *prev, *next;

   ASSERT(hint != NULL);
   prev = hint->prevByKey;
   next = hint->nextByKey;
   if (prev != NULL) {
      prev->nextByKey = next;
   } else {
      *Vmx86COWHintTableKeyChainHeadPtr(hint->key) = next;
   }
   if (next != NULL) {
      next->prevByKey = prev;
   } 
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintTableMPNRemove --
 *
 *       Unlinks COW hint frame from the hash table that keeps track of 
 *       page numbers. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       Element specified is removed from the corresponding COW hint hash 
 *       table.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86COWHintTableMPNRemove(COWHint *hint)
{
   COWHint *prev, *next;

   ASSERT(hint != NULL);
   prev = hint->prevByMPN;
   next = hint->nextByMPN;
   if (prev != NULL) {
      prev->nextByMPN = next;
   } else {
      *Vmx86COWHintTableMPNChainHeadPtr(hint->origMPN) = next;
   }
   if (next != NULL) {
      next->prevByMPN = prev;
   } 
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWHintStorageRemove --
 *
 *       Unlinks COW hint frame from both hash tables. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       Element specified is removed from both COW hint hash tables.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86COWHintStorageRemove(COWHint *hint)
{
   Vmx86COWHintTableKeyRemove(hint);
   Vmx86COWHintTableMPNRemove(hint);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86LockedPageLimit --
 *
 *       There are three limits controlling how many pages we can lock on 
 *       a host:  
 *
 *       lockedPageLimit.configured is controlled by UI,  
 *       lockedPageLimit.dynamic is controlled by authd's hardLimitMonitor,
 *       lockedPageLimit.host is calculated dynamically based on kernel stats 
 *       by vmmon using kernel stats.
 *
 *       We can lock the MIN of these values.
 *
 * Results:
 *       Number of pages to lock on this host.
 *
 * Side effects:
 *       Updates the host locked pages limit.
 *
 *----------------------------------------------------------------------
 */

static INLINE unsigned
Vmx86LockedPageLimit(const VMDriver* vm)
{
   ASSERT(HostIF_GlobalLockIsHeld());

   lockedPageLimit.host = HostIF_EstimateLockedPageLimit(vm, numLockedPages);
 
   return MIN(MIN(lockedPageLimit.configured, lockedPageLimit.dynamic),
              lockedPageLimit.host);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86LockedPageLimitForAdmissonControl --
 *
 *       There are two limits controlling how many pages we can lock on 
 *       a host:  
 *
 *       lockedPageLimit.configured is controlled by UI,  
 *       lockedPageLimit.dynamic is controled by authd's hardLimitMonitor,
 *
 *       We can lock the MIN of these values.
 *
 *       Using lockedPageLimit.host would be too pessimistic.  After admission
 *       of a new VM but before allocation/locking memory for the new VM 
 *       our memory sharing code will put pressure on other VMs and should 
 *       produce enough free pages to successfully finish poweron.
 *
 * Results:
 *       Number of pages to lock on this host.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER unsigned
Vmx86LockedPageLimitForAdmissonControl(void)
{
   ASSERT(HostIF_GlobalLockIsHeld());
   return MIN(lockedPageLimit.configured, lockedPageLimit.dynamic);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86HasFreePages --
 *
 *       Returns TRUE if the vm can lock more pages.  This is true if 
 *       we are below the host's hard memory limit and this vm has not
 *       exceeded its maximum allocation.
 *       Callers must ensure driver-wide and VM serialization
 *       typically by using HostIF_GlobalLock() and  HostIF_VMLock().
 *
 * Results:
 *       TRUE if pages can be locked, FALSE otherwise
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
Vmx86HasFreePages(VMDriver *vm, 
		  unsigned int numPages,
                  Bool checkVM)
{
   /*  
    * 1) Be careful with overflow. 
    * 2) lockedPageLimit and vm->memInfo.maxAllocation can be decreased below
    *    the current numLockedPages and vm->memInfo.locked
    * 3) lockedPageLimit.host can go lower than numLockedPages.
    */
   ASSERT(HostIF_GlobalLockIsHeld() &&
          (!checkVM || HostIF_VMLockIsHeld(vm)));

   if (checkVM) {
      /*
       * Check the per-vm limit.
       */
      ASSERT(HostIF_VMLockIsHeld(vm));
      if (vm->memInfo.admitted) {
	 if (vm->memInfo.maxAllocation <= vm->memInfo.locked) {
	    return FALSE;
	 } else if (vm->memInfo.maxAllocation - vm->memInfo.locked < numPages) {
	    return FALSE;
	 }
      }
   } else {
      /*
       * Check the global limit.
       */
      unsigned limit = Vmx86LockedPageLimit(vm);

      if (limit <= numLockedPages) {
	 return FALSE;
      } else if (limit - numLockedPages < numPages) {
	 return FALSE;
      }
   }
   return TRUE;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86CowFreeID --
 *
 *       Mark reference count location id as unused in the COWPage ref
 *       count array.  
 *
 * Results:
 *       Clears ref count for all pages by vm with id.  
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86COWFreeID(int id)
{ 
   int i,tmp;

   ASSERT(HostIF_GlobalLockIsHeld());

   /* deleting head of the list */
   if (id == cowIDsAllocated) {
      tmp = cowIDList[cowIDsAllocated];
      cowIDList[cowIDsAllocated] = cowIDsUnused;
      cowIDsAllocated = tmp;
      cowIDsUnused = id;
      return;
   }
      
   for (i = cowIDsAllocated; cowIDList[i] != INVALID_COWID; i = cowIDList[i]) {
      if (cowIDList[i] == id) {
         cowIDList[i] = cowIDList[id];
         cowIDList[id] = cowIDsUnused;
         cowIDsUnused = id;
         return;
      }
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWAllocID --
 *
 *       Mark reference count location id as used in the COWPage ref
 *       count array.  
 *
 * Results:
 *       Allows vm to use returned index when ref counting COW pages.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static int 
Vmx86COWAllocID(void)
{
   int i;
   ASSERT(HostIF_GlobalLockIsHeld());
   i = cowIDsUnused;
   ASSERT(i != INVALID_COWID);
   cowIDsUnused = cowIDList[i];
   cowIDList[i] = cowIDsAllocated;
   cowIDsAllocated = i;
   return i;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageRef --
 *
 *       Determines the total ref count for a COW page.
 *
 * Results:
 *       ref count for all valid vms for this COW page
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static uint32
Vmx86COWPageRef(COWPage *entry)
{
   int i;
   uint32 sum = 0;
   
   if (entry != NULL) {
      for (i = cowIDsAllocated; i != INVALID_COWID; i = cowIDList[i]) {
         sum += entry->ref[i];
      }
   }
   return sum;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageIsStale --
 *
 *       Determines if COW page has a zero ref count and is NOT locked
 *       as a shared page.
 *
 * Results:
 *       TRUE if entry is referenced, FALSE otherwise.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
Vmx86COWPageIsStale(COWPage *entry)
{
   ASSERT(entry != NULL);
   return Vmx86COWPageRef(entry) == 0 && !entry->locked;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageIsReferenced --
 *
 *       Determines if COW page has a non-zero ref count for some
 *       running virtual machine.
 *
 * Results:
 *       TRUE if entry is referenced, FALSE otherwise.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
Vmx86COWPageIsReferenced(COWPage *entry)
{
   ASSERT(entry != NULL);
   return Vmx86COWPageRef(entry) != 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWAllocHint --
 *
 *       Create a COW hint frame and link it to the COW hint hash table.  
 *
 * Results:
 *       Returns pointer to allocated COW hint frame on success, otherwise
 *       returns NULL.
 *
 * Side effects:
 *       Allocates kernel memory.
 *
 *----------------------------------------------------------------------
 */

static COWHint *
Vmx86COWAllocHint(VMDriver *vm, BPN bpn, uint64 key, MPN mpn)
{
   COWHint *entry; 

   ASSERT(HostIF_GlobalLockIsHeld());
   if (cowState.numHints >= COW_MAX_HINTS) {
      return NULL;
   }
   entry = HostIF_AllocKernelMem(sizeof *entry, FALSE);
   if (entry != NULL) {
      entry->vm = vm;
      entry->bpn = bpn;
      entry->origMPN = mpn;
      entry->key = key;
      Vmx86COWHintStorageAdd(entry);
      cowState.numHints++;
   }
   return entry;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWFreeHint--
 *
 *       Deallocate a COW hint frame pointed to by hint. 
 *
 * Results:
 *       None
 *
 * Side effects:
 *       Hint is deleted from hash table, kernel memory is returned.
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86COWFreeHint(COWHint *hint)
{
   ASSERT(HostIF_GlobalLockIsHeld());
   Vmx86COWHintStorageRemove(hint);
   memset(hint, 0, sizeof (COWHint));
   HostIF_FreeKernelMem(hint);
   cowState.numHints--;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWAllocPage--
 *
 *       Allocate a COW page and create a frame for it. Link the frame to
 *       COW page hash table.
 *
 * Results:
 *       Returns pointer to allocated COW page frame on success, otherwise
 *       returns NULL.
 *
 * Side effects:
 *       Allocates kernel memory.
 *
 *----------------------------------------------------------------------
 */

static COWPage *
Vmx86COWAllocPage(VMDriver *vm, uint64 key, char *buf, Bool locked)
{
   COWPage *entry = NULL;
   MPN mpn;
   ASSERT(HostIF_GlobalLockIsHeld());
   /*
    * We limit the number of unique MPNs for COW to 20%
    * of host memory.  Since we use an extra MPN for cow pages
    * this limit is important to avoid swapping the host to 
    * death if the vm decides to break cow on all its pages.
    */
   if (((cowState.uniqueMPNs < Vmx86LockedPageLimit(vm) / 5 &&
         Vmx86HasFreePages(vm, 1, FALSE)) || vm == NULL) &&
       (mpn = HostIF_COWAllocPage()) != INVALID_MPN) {

      entry = HostIF_AllocKernelMem(sizeof *entry, FALSE);
      if (entry != NULL) {
         if (HostIF_WritePage(mpn, buf, TRUE) == 0) {
            memset((char *)entry->ref, 0, sizeof(entry->ref[0]) * MAX_VMS);
            entry->sharedMPN = mpn;
            entry->locked = locked;
            entry->key = key;
            Vmx86COWPageTableAdd(entry);
            cowState.uniqueMPNs++;      
            cowState.totalUniqueMPNs++;
            numLockedPages++;
         } else {
            HostIF_COWFreePage(mpn);
            memset(entry, 0, sizeof (COWPage));
            HostIF_FreeKernelMem(entry);
	    entry = NULL;
         }
      } else {
         HostIF_COWFreePage(mpn);
      }
   }
   return entry;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWFreePage--
 *
 *       Deallocate both COW page and COW page frame. Frame pointed to 
 *       by 'prev' must precede the one pointed to by 'page' in a hash
 *       chain.
 *
 * Results:
 *       None
 *
 * Side effects:
 *       COW page frame is deleted from hash table, kernel memory is 
 *       released.
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86COWFreePage(COWPage *page, COWPage *prev)
{
   ASSERT(HostIF_GlobalLockIsHeld());
   ASSERT(!Vmx86COWPageIsReferenced(page));
   ASSERT(numLockedPages > 0);

   HostIF_COWFreePage(page->sharedMPN);
   numLockedPages--;

   Vmx86COWPageTableRemove(page, prev);
   memset(page, 0, sizeof (COWPage));
   HostIF_FreeKernelMem(page);
   cowState.uniqueMPNs--;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWGetPageClassID --
 *
 *       Determines if a page contains some regular pattern on it.
 *       In a trivial case looks for zero contents. Optionally may
 *       look for other patterns which might help to recognize good
 *       candidates for sharing.
 *
 * Results:
 *       Returns identifier of a class to which the page belongs.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER PageClassID
Vmx86COWGetPageClassID(const char *buf)
{
   uint32 *p = (uint32*) buf;
   uint32 i = 0, n = PAGE_SIZE / sizeof(uint32);
   while (p[i] == 0 && ++i < n);
   return (i == n)? PCID_ZERO : PCID_UNKNOWN;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWGetPageSignature --
 *
 *       Computes a signature for a given machine page which is either a 
 *       hash of its contents or an identifier of a well-defined class 
 *       to which the page belongs.
 *
 * Results:
 *       Reads mpn into buffer provided and computes page signature for
 *       use in COW. Returns 0 on success.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */
static INLINE_SINGLE_CALLER int
Vmx86COWGetPageSignature(MPN mpn, char *buf, PageSig *ps, Bool *isID)
{
   int retval;
   retval = HostIF_ReadPage(mpn, buf, TRUE);
   if (retval == 0) {
      if (Vmx86COWGetPageClassID(buf) == PCID_ZERO) {
         ps->id = PCID_ZERO;
         *isID = TRUE;
      } else {
         ps->key = Hash_Page(buf);
         *isID = FALSE;
      }
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWReadMPNIntoBuf --
 *
 *       Reads mpn into buffer provided and returns TRUE on success.
 *
 * Results:
 *       Returns TRUE on success, FALSE otherwise.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
Vmx86COWReadMPNIntoBuf(MPN mpn,      // IN
                       char *buf)    // OUT
{
   return HostIF_ReadPage(mpn, buf, TRUE) == 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWStats --
 *
 *       Dump stats about page sharing.  We identify the pages with
 *       the highest ref counts and report them to the vmx.
 *
 * Results:
 *       COW sharing statistics are filled into info.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86COWStats(VMMemCOWInfo *info)
{
   COWPage *hot[VMMEM_COW_HOT_PAGES];
   COWPage *entry;
   COWPage *tmp;
   int i;
   int j,k;
   uint32 totalRef = 0;


   for (i = 0; i < VMMEM_COW_HOT_PAGES; i++) {
      hot[i] = NULL;
   }
   
   for (i = 0; i < COW_PAGE_TABLE_SIZE; i++) {
      entry = cowPageTable[i];
      while (entry != NULL) {
         if (Vmx86COWPageRef(entry) > Vmx86COWPageRef(hot[0])) {
            hot[0] = entry;
            for (j = VMMEM_COW_HOT_PAGES - 1; j >= 0; j--) {
               for (k = 0; k < j; k++) {
                  if (Vmx86COWPageRef(hot[k]) > Vmx86COWPageRef(hot[k + 1])) {
                     tmp = hot[k];
                     hot[k] = hot[k + 1];
                     hot[k + 1] = tmp;
                  }
               }
            }
         }
         totalRef += Vmx86COWPageRef(entry);
         entry = entry->next;
      }
   }
   for (i = 0; i < VMMEM_COW_HOT_PAGES; i++) {
      if (hot[i] != NULL) {
         info->hot[i].mpn = hot[i]->sharedMPN;
         info->hot[i].key = hot[i]->key;
         info->hot[i].ref = Vmx86COWPageRef(hot[i]);
      } else {
         info->hot[i].mpn = INVALID_MPN;
         info->hot[i].ref = 0;
         info->hot[i].key = 0;
      }
   }
   info->numRef = totalRef;
   info->numHints = cowState.numHints;
   info->uniqueMPNs = cowState.uniqueMPNs;
   info->numBreaks = cowState.numBreaks;
   info->totalUniqueMPNs = cowState.totalUniqueMPNs;
}

   
/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWPageTableWalk --
 *
 *       Walk the table of COW page frames looking for key.
 *
 * Results:
 *       Returns TRUE if hash is found, otherwise returns FALSE. Pointers 
 *       prev and entry are set correctly.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static Bool
Vmx86COWPageTableWalk(uint64 key, COWPage **entry, COWPage **prev, 
                      COWPageWalkArgs args)
{
   ASSERT(HostIF_GlobalLockIsHeld());
   *entry = *Vmx86COWPageTableChainHeadPtr(key);
   *prev = NULL;
   while (*entry != NULL) {
      if ((args == COWMatchKey && (*entry)->key == key) || 
          (args == COWMatchStale && Vmx86COWPageIsStale(*entry)) ||
          (args == COWMatchLocked && (*entry)->locked)) {
         return TRUE;
      }
      *prev = *entry;
      *entry = (*entry)->next;
   }
   return FALSE;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWLookupHintByKey --
 *
 *       Walk the table of COW hint frames looking for key.
 *
 * Results:
 *       Returns pointer to hint if a hint frame holding the key is found,
 *       otherwise returns NULL.
 *
 * Side effects:
 *       If hint is found the lock for the VM that owns the hint is held
 *       on return.
 *
 *----------------------------------------------------------------------
 */

static COWHint*
Vmx86COWLookupHintByKey(uint64 key, int callerID)
{
   COWHint *hint;
   ASSERT(HostIF_GlobalLockIsHeld());
   hint = *Vmx86COWHintTableKeyChainHeadPtr(key);

   while (hint != NULL) {
      VMDriver *hintVM = hint->vm;
      HostIF_VMLock(hintVM, callerID);
      if (!HostIF_IsLockedByMPN(hintVM, hint->origMPN)) {
         COWHint *next = hint->nextByKey;
         Vmx86COWFreeHint(hint);
         hint = next;
      } else {
         if (hint->key == key) {
            return hint;
         }
         hint = hint->nextByKey;
      }
      HostIF_VMUnlock(hintVM, callerID);
   }
   return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWLookupHintByMPN --
 *
 *       Walk the table of COW hint frames looking for Page Number.
 *
 * Results:
 *       Returns pointer to hint if a hint frame for the page is found,
 *       otherwise returns NULL.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static COWHint*
Vmx86COWLookupHintByMPN(MPN mpn)
{
   COWHint *hint;
   ASSERT(HostIF_GlobalLockIsHeld());
   hint = *Vmx86COWHintTableMPNChainHeadPtr(mpn);

   while (hint != NULL) {
      if (!HostIF_IsLockedByMPN(hint->vm, hint->origMPN)) {
         COWHint *next = hint->nextByMPN;
         Vmx86COWFreeHint(hint);
         hint = next;
      } else {
         if (hint->origMPN == mpn) {
            return hint;
         }
         hint = hint->nextByMPN;
      }
   }
   return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_COWRemoveHint --
 *
 *       Searches for the COW hint for this mpn and removes it from 
 *       the hint table.
 *
 * Results:
 *       None
 *
 * Side effects:
 *       Hint is removed.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_COWRemoveHint(VMDriver *vm, MPN mpn)
{
   COWHint *hint;

   HostIF_GlobalLock(28);
   if ((hint = Vmx86COWLookupHintByMPN(mpn)) != NULL) {
      Vmx86COWFreeHint(hint);
   }
   HostIF_GlobalUnlock(28);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWSharePageSlow --
 *
 *       Attempt to share guest page "bpnToShare" which corresponds
 *       to machine page mpn. We first try to share with an already
 *       existing COW page. If no matching page exists, we check
 *       the hint table for a potential sharing opportunity. If 
 *       no matching hint exists, we add a hint for this page for
 *       future sharing opportunities.  
 *
 * Results:
 *       Page may be shared, hint page may be added. BpnToShare is
 *       INVALID_BPN if no sharing is possible and a hint cannot
 *       be added. HintOnly Bool is TRUE iff a hint is added for this
 *       page. MPN contains the potentially new MPN corresponding
 *       guest page "bpnToShare".  
 *
 * Side effects:
 *       If a page was shared, updates the corresponding entry in the 
 *       table of COW pages. May add/remove entries to/from the table
 *       of COW hint.
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86COWSharePageSlow(VMDriver *vm, BPN *bpnToShare, MPN *mpn,
                      Bool *hintOnly, uint64 key)
{
   COWPage *entry, *prev;
   COWHint *hint;
   BPN bpn = *bpnToShare;
   PShare_HintUpdate hintUpdate;
   VMDriver *hintVM;
   Bool doAllocHint = FALSE;

   ASSERT(HostIF_GlobalLockIsHeld());
   
   *bpnToShare = INVALID_BPN;
   *hintOnly = FALSE;
   
   /*
    * Search among hints to match mpn. We do not want to keep multiple
    * hints for the same machine page because only one of them is valid.
    * Also we do not want to do anything if a hint is installed for the
    * mpn and this hint is in good standing.
    */
   if ((hint = Vmx86COWLookupHintByMPN(*mpn)) != NULL) {
      hintVM = hint->vm;
      if (hint->key != key) {
         /*
          * If contents of the page has been changed remove the hint to try
          * matching new page contents later on. We should not bother with 
          * posting updates if the hint was originally installed for another
          * vm/bpn. What happened is that another/this VM unlocked the page,
          * swapped it out, and this VM swapped it in. At unlock time hint
          * bit was already cleared for another VM/bpn.
          */
         Vmx86COWFreeHint(hint);
      } else {
         /*
          * If contents of the page has not been changed it is enough to 
          * update vm/bpn fields in the hint frame.
          */
         hint->bpn = bpn;
         hint->vm = vm;
         *hintOnly = TRUE;
         *bpnToShare = bpn;
         return;
      }
   }

   if (Vmx86COWPageTableWalk(key, &entry, &prev, COWMatchKey)) {
      /*
       * Match in the COW page table. Do a full comparison and
       * share the page. If a false match occurs, we abandon this
       * page altogether without trying to share it further.  
       */
      ASSERT(entry->sharedMPN != INVALID_MPN);
      if (HostIF_ReadPage(entry->sharedMPN, cowBuf, TRUE) == 0 &&
          HostIF_ReadPage(*mpn, cowContents[vm->cowID], TRUE) == 0 &&
	  memcmp(cowBuf, cowContents[vm->cowID], PAGE_SIZE) == 0) {
         HostIF_VMLock(vm, 24);
         entry->ref[vm->cowID]++;
         vm->memInfo.shared++;
         HostIF_VMUnlock(vm, 24);
         *mpn = entry->sharedMPN;
         *bpnToShare = bpn;
      }
   } else if ((hint = Vmx86COWLookupHintByKey(key, 27)) != NULL) {
      /*
       * Match in the hint table. Refresh the hint hash to make sure
       * it is still valid. If it is, create a shared page and queue
       * an update to the hint owner to try sharing the page again.
       *
       * We only need to keep the origMPN until its contents have
       * been copied into "cowBuf". After that, we only interact with
       * hint structures which are guarded by the global lock or
       * we get the appropriate vm-lock as needed.
       */
      Bool success;

      hintVM = hint->vm;
      ASSERT(HostIF_IsLockedByMPN(hintVM, hint->origMPN));
      success = Vmx86COWReadMPNIntoBuf(hint->origMPN, cowBuf);
      HostIF_VMUnlock(hintVM, 27);

      if (success) {
         if (hint->bpn == bpn && hintVM == vm) {
            /*
             * Hint frame has to be rehashed since at least mpn has been 
             * changed for it. This is the case when a machine page to which 
             * the guest page was mapped has been swapped out and then swapped
             * in to a different machine page shortly so that hint was not 
             * removed as the one referencing to an unlocked machine page.
             * After swapping in the contents might have also been changed.
             *
             * TODO: It is sad that hint has to be reallocated just after it
             * was freed. We can save if we just reuse that memory.
             */
            Vmx86COWFreeHint(hint);
            doAllocHint = TRUE;
         } else {
            if (hintVM->hintIndex < PSHARE_HINT_UPDATES_MAX) {
               hintUpdate.bpn = hint->bpn;
               if (Hash_Page(cowBuf) != hint->key) {
                  // TODO: Free hint right away and post PSHARE_HINT_NONE
                  hintUpdate.status = PSHARE_HINT_STALE;
                  HostIF_VMLock(hintVM, 25);
                  hintVM->hintUpdate[hintVM->hintIndex++] = hintUpdate;
                  HostIF_VMUnlock(hintVM, 25);
                  doAllocHint = TRUE;
               } else {
                  char *buf = cowContents[vm->cowID];
                  if (HostIF_ReadPage(*mpn, buf, TRUE) == 0) {
                     entry = Vmx86COWAllocPage(vm, key, buf, FALSE);
                     if (entry != NULL) {
                        Vmx86COWFreeHint(hint);

                        HostIF_VMLock(vm, 28);
                        entry->ref[vm->cowID]++;
                        vm->memInfo.shared++;
                        HostIF_VMUnlock(vm, 28);

                        *mpn = entry->sharedMPN;
                        *bpnToShare = bpn;

                        hintUpdate.status = PSHARE_HINT_MATCH;
                        HostIF_VMLock(hintVM, 29);
                        hintVM->hintUpdate[hintVM->hintIndex++] = hintUpdate;
                        HostIF_VMUnlock(hintVM, 29);
                     }
                  }
               }
            }
         }
      }
   } else {
      /* 
       * No match in either the hint table or page table.
       * Add a hint for this page for future sharing.
       */
      doAllocHint = TRUE;
   }

   if (doAllocHint) {
      if (Vmx86COWAllocHint(vm, bpn, key, *mpn) != NULL) {
         *hintOnly = TRUE;
         *bpnToShare = bpn;
      }
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86COWFreeAllResources --
 *
 *      Clean up COW state associated with this vm.  
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Free all hints and pages associated with vm. 
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86COWFreeAllResources(VMDriver *vm)
{
   COWHint *hint;
   COWPage *entry, *prev;
   int i;

   if (vm->cowID != INVALID_COWID) {
      ASSERT(HostIF_GlobalLockIsHeld());     
      for (i = 0; i < COW_HINT_TABLE_KEY_SIZE; i++) {
         hint = *Vmx86COWHintTableKeyChainHeadPtr(i);
         while (hint != NULL) {
            if (vm == hint->vm) {
               COWHint *next = hint->nextByKey;
               Vmx86COWFreeHint(hint);
               hint = next;
            } else {
               hint = hint->nextByKey;
            }
         }
      }

      /*
       * We could reuse the cowID for this vm so we need to explicitly clear
       * the ref counts in the hash table.
       */
      for (i = 0; i < COW_PAGE_TABLE_SIZE; i++) {
         entry = *Vmx86COWPageTableChainHeadPtr(i);
         while (entry != NULL) {
            entry->ref[vm->cowID] = 0;
            entry = entry->next;
         }
      }

      /*
       * Now we can reap unreferenced pages. 
       */
      for (i = 0; i < COW_PAGE_TABLE_SIZE; i++) {
         while(Vmx86COWPageTableWalk(i, &entry, &prev, COWMatchStale)) {
            Vmx86COWFreePage(entry, prev);
         }
      }
   }
}


#ifdef VMX86_DEBUG

/*
 *----------------------------------------------------------------------
 *
 * Vmx86VMIsRegistered --
 *
 *      Check if "vm" is on the list of VMDrivers.
 *
 * Results:
 *      Return TRUE iff "vm" is on the list of VMDrivers.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------
 */

static Bool
Vmx86VMIsRegistered(VMDriver *vm, Bool needsLock) 
{
   VMDriver *tmp;
   Bool      found = FALSE;

   ASSERT(needsLock || HostIF_GlobalLockIsHeld());

   if (needsLock) {
      HostIF_GlobalLock(5);
   }

   for (tmp = vmDriverList; tmp != NULL; tmp = tmp->nextDriver) {
      if (tmp == vm) {
         found = TRUE;
         break;
      }
   }

   if (needsLock) {
      HostIF_GlobalUnlock(5);
   }

   return found;
}

#endif

/*
 *----------------------------------------------------------------------
 *
 * Vmx86RegisterVMOnList --
 *
 *      Add a VM to the list of registered VMs and increment
 *      the count of VMs.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Add VM to linked list.
 *	Increment count of VMs.
 *
 *----------------------------------------------------------------
 */

static void
Vmx86RegisterVMOnList(VMDriver *vm) 
{
   VMDriver **vmp;

   ASSERT(HostIF_GlobalLockIsHeld());
   vmCount++;
   ASSERT(vm->cowID == INVALID_COWID);
   vm->cowID = Vmx86COWAllocID();
   vm->count++;
   for (vmp = &vmDriverList; *vmp != NULL; vmp = &(*vmp)->nextDriver) {
      if (*vmp == vm) {
         Warning("VM %p already registered on the list of VMs.\n", vm);
         return;
      }
   }
   *vmp = vm;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86DeleteVMFromList --
 *
 *      Delete a VM from the list of registered VMs and decrement
 *      the count of VMs. This function should be called on any
 *      VM registered on the VMDriverList before invoking 
 *      Vmx86FreeAllVMResources to free its memory.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Remove VM from linked list.
 *	Decrement count of VMs.
 *
 *----------------------------------------------------------------
 */

static void
Vmx86DeleteVMFromList(VMDriver *vm) 
{
   VMDriver **vmp;

   ASSERT(HostIF_GlobalLockIsHeld());
   for (vmp = &vmDriverList; *vmp != vm; vmp = &(*vmp)->nextDriver) {
      if (*vmp == NULL) {
         Warning("VM %p is not on the list of registered VMs.\n", vm);
         return;
      }
   }
   *vmp = vm->nextDriver;
   vmCount--;

   Vmx86COWFreeAllResources(vm);
   Vmx86COWFreeID(vm->cowID);   
   numLockedPages -= vm->memInfo.locked;

   /*
    * If no VM is running, reset the configured locked-page limit so
    * that the next VM to power on sets it appropriately.
    */
   if (vmCount == 0) {
      lockedPageLimit.configured = MAX_LOCKED_PAGES; /* cannot do 0, WS4.5.2 sets limit after initial admission */
   }
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86FreeAllVMResources
 *
 *     Free the resources allocated for a vm that is not registered
 *     on the VMDriverList.  Except in the case of Vmx86_CreateVM(), 
 *     this should be called only after a call to Vmx86DeleteVMFromList().
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Memory freed.
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86FreeAllVMResources(VMDriver *vm)
{
   ASSERT(!HostIF_GlobalLockIsHeld());
   if (vm) {   
      ASSERT(!Vmx86VMIsRegistered(vm, TRUE));

#ifdef USE_PERFCTRS_HOSTED
      /*
       * Can't log in PerfCtr_Release() when we're holding the lock
       * and (in any case) the VM has be dequeued.
       */
      PerfCtr_Release(vm, FALSE);
#endif

      Vmx86_SetHostClockRate(vm, 0);

      HostIF_FreeAllResources(vm);

      if (vm->vmhost != NULL) {
         HostIF_FreeKernelMem(vm->vmhost);
      }
      HostIF_FreeKernelMem(vm);
   }
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_COWCopyPage --
 *
 *       Break COW sharing on sharedMPN.  If sharedMPN is no longer
 *       referenced, it is returned to the host.
 *
 * Results:
 *       None
 *
 * Side effects:
 *       Decrement ref count on sharedMPN and potentially free the 
 *       shared page.     
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_COWCopyPage(VMDriver *vm, MPN sharedMPN)
{
   COWPage *entry, *prev;

   HostIF_GlobalLock(26);
   HostIF_VMLock(vm, 26);
   ASSERT(cowZeroPage != NULL);
   if (sharedMPN == cowZeroPage->sharedMPN) {
      /*
       * Make sure that breaking COW sharing for the zero page
       * does not involve hash computation.
       */
      ASSERT(cowZeroPage->ref[vm->cowID] > 0);
      cowZeroPage->ref[vm->cowID]--;
      vm->memInfo.shared--;
      cowState.numBreaks++;
   } else {
      if (Vmx86COWReadMPNIntoBuf(sharedMPN, cowBuf)) {
         if (Vmx86COWPageTableWalk(Hash_Page(cowBuf), &entry, &prev, COWMatchKey)) {
            if (sharedMPN == entry->sharedMPN) {
               ASSERT(entry->ref[vm->cowID] > 0);
               entry->ref[vm->cowID]--;
               vm->memInfo.shared--;
               cowState.numBreaks++;
               if (Vmx86COWPageIsStale(entry)) {
                  Vmx86COWFreePage(entry, prev);
               }
            }
         }
      }
   }
   HostIF_VMUnlock(vm, 26);
   HostIF_GlobalUnlock(26);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_COWSharePages --
 *
 *      Attempt to share page list described by info.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      VCPU thread continues.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_COWSharePages(VMDriver *vm, COWShareInfo *info)
{
   static char cowList[MAX_VMS][PAGE_SIZE];
   static PageSig cowPageSig[MAX_VMS][PSHARE_BATCH_PAGES_MAX];
   static Bool cowIsID[MAX_VMS][PSHARE_BATCH_PAGES_MAX];

   if (HostIF_ReadPage(info->pshareMPN, cowList[vm->cowID], TRUE) == 0) {
      uint32 i, zeroCnt = 0;
      BusMemDef_BatchPageList *list;
      char *buf = cowContents[vm->cowID];
      PageSig *ps = cowPageSig[vm->cowID];
      Bool *isID = cowIsID[vm->cowID];

      list = (BusMemDef_BatchPageList *)cowList[vm->cowID];
      for (i = 0; i < info->numPages; i++) {
         if (Vmx86COWGetPageSignature(list->mpnList[i], buf, 
                                      &ps[i], &isID[i]) != 0) {
            list->bpnList[i] = INVALID_BPN;
         }
      }

      ASSERT(cowZeroPage != NULL);
      HostIF_GlobalLock(25);
      for (i = 0; i < info->numPages; i++) {
         if (list->bpnList[i] != INVALID_BPN) {
            if (isID[i]) {
               COWHint *hint;
               ASSERT(ps[i].id == PCID_ZERO);
               if ((hint = Vmx86COWLookupHintByMPN(list->mpnList[i])) != NULL){
                  // no hint update required (see Vmx86COWSharePage)
                  Vmx86COWFreeHint(hint);
               }
               list->mpnList[i] = cowZeroPage->sharedMPN;
               list->flagList[i] = FALSE;
               zeroCnt++;
            } else {
               Vmx86COWSharePageSlow(vm, &list->bpnList[i], &list->mpnList[i],
                                     &list->flagList[i], ps[i].key);
            }
         }
      }
      HostIF_VMLock(vm, 30);
      cowZeroPage->ref[vm->cowID] += zeroCnt;
      vm->memInfo.shared += zeroCnt;
      HostIF_VMUnlock(vm, 30);
      HostIF_GlobalUnlock(25);

      if (HostIF_WritePage(info->pshareMPN, cowList[vm->cowID], TRUE) != 0) {
         info->shareFailure = TRUE;
      }
   } else {
      info->shareFailure = TRUE;
   }

   // hints waiting for us from other vms?
   info->updateHints = vm->hintIndex > 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_COWGetHintUpdates --
 *
 *      Attempt to get hint updates.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Moon turns to cheese.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_COWGetHintUpdates(VMDriver *vm, COWHintInfo *info)
{
   unsigned numHints;
   ASSERT(info->nUpdates <= PSHARE_HINT_BATCH_PAGES_MAX);
   HostIF_VMLock(vm, 23);
   numHints = MIN(vm->hintIndex, info->nUpdates);
   memcpy(info->updates, vm->hintUpdate + vm->hintIndex - numHints, 
          numHints * sizeof(PShare_HintUpdate));
   vm->hintIndex = vm->hintIndex - numHints;
   info->nUpdates = numHints;
   HostIF_VMUnlock(vm, 23);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_COWCheckPages --
 *
 *      Verifies the consistency of memory pages.  This allows the monitor
 *      react to potential memory corruption or a disagreement in cow state
 *      between the vmm and vmmon.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Warnings/messages could be spewed if memory corruption is found.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_COWCheckPages(VMDriver *vm, COWCheckInfo *info)
{
   unsigned i;

   for(i = 0; i < info->numPages; i++) {
      COWPage *entry, *prev;
      PShare_COWCheckInfo *check = &info->check[i];
      check->checkOK = TRUE;
      HostIF_GlobalLock(29);
      if (Vmx86COWReadMPNIntoBuf(check->vmmMPN, cowBuf) &&
          Vmx86COWPageTableWalk(Hash_Page(cowBuf), &entry, &prev, COWMatchKey)) {
         if (entry->sharedMPN == check->vmmMPN &&
            entry->ref[vm->cowID] > 0) {
            check->hostCOW = TRUE;
            check->hostMPN = entry->sharedMPN;
            if (!check->vmmCOW) {
               Warning("Bad COW state: bpn=%#x mpn=%#x "
                       "looks non-shared in VMM\n", check->bpn, check->vmmMPN);
            }
         } else {
            check->hostCOW = FALSE;
            check->hostMPN = check->vmmMPN;
            if (check->vmmCOW) {
               Warning("Bad COW state: bpn=%#x mpn=%#x "
                       "looks shared in VMM\n", check->bpn, check->vmmMPN);
            }
         }
      } else {
         check->hostCOW = FALSE;
         check->hostMPN = check->vmmMPN;
         if (check->vmmCOW) {
            Warning("Bad COW state: bpn=%#x mpn=%#x "
                    "looks shared in VMM\n", check->bpn, check->vmmMPN);
         }
      }
      HostIF_GlobalUnlock(29);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_COWGetZeroMPN --
 *
 *      Retrieve shared mpn for the zero page.  
 *
 * Results:
 *      Single zero mpn.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

MPN
Vmx86_COWGetZeroMPN(void)
{
   ASSERT(cowZeroPage != NULL);
   return cowZeroPage->sharedMPN;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_InitCOWList --
 *
 *       Set up the list of remaining cowID's. These index into the ref
 *       count field of the cow page structure. Preallocate zero COW page.
 *       Called at insmod time.
 *
 * Results:
 *       Sets up global data.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_InitCOWList(void)
{
   int i;
   uint64 cowZeroKey;

   HostIF_GlobalLock(32);
   for (i = 0; i < MAX_VMS; i++) {
      cowIDList[i] = i + 1;
   }
   cowIDList[MAX_VMS - 1] = INVALID_COWID;
   cowIDsUnused = 0;
   cowIDsAllocated = INVALID_COWID;

   /*
    * Always keep a locked reference to the zero page.
    */
   memset(cowBuf, 0, PAGE_SIZE);
   cowZeroKey = Hash_Page(cowBuf);
   cowZeroPage = Vmx86COWAllocPage(NULL, cowZeroKey, cowBuf, TRUE);
   HostIF_GlobalUnlock(32);
}


void
Vmx86_COWCleanup(void)
{
   COWPage *entry, *prev;
   int i;
   HostIF_GlobalLock(34);
   for (i = 0; i < COW_PAGE_TABLE_SIZE; i++) {
      while(Vmx86COWPageTableWalk(i, &entry, &prev, COWMatchLocked)) {
         Vmx86COWFreePage(entry, prev);
      }
   }

   // all entries should be gone now!
   for (i = 0; i < COW_PAGE_TABLE_SIZE; i++) {
      ASSERT(*Vmx86COWPageTableChainHeadPtr(i) == NULL);
   }
   HostIF_GlobalUnlock(34);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86ReserveFreePages --
 *
 *       Returns TRUE and increases locked page counts if the vm can lock 
 *       more pages.  This is true if we are below the host's hard memory 
 *       limit and this vm has not exceeded its maximum allocation.
 *       The function is thread-safe.
 *
 * Results:
 *       TRUE if pages are reserved for locking, FALSE otherwise
 *
 * Side effects:
 *       The global lock and VM's lock are acquired and released.
 *
 *----------------------------------------------------------------------
 */

static Bool
Vmx86ReserveFreePages(VMDriver *vm, 
		      unsigned int numPages)
{
   Bool retval = FALSE;
   int retries = 3;

   ASSERT(vm);
   
   for (retries = 3; !retval && (retries > 0); retries--) {
      HostIF_GlobalLock(17);
      HostIF_VMLock(vm, 0);
      
      // Check VM's limit and don't wait.
      retval = Vmx86HasFreePages(vm, numPages, TRUE);
      if (!retval) {
         HostIF_VMUnlock(vm, 0);
         HostIF_GlobalUnlock(17);
	 break;
      } else {
	 // Wait to satisfy the global limit.
	 retval = Vmx86HasFreePages(vm, numPages, FALSE);
	 if (retval) {
	    numLockedPages += numPages;
	    vm->memInfo.locked += numPages;
            HostIF_VMUnlock(vm, 0);
	    HostIF_GlobalUnlock(17);
	    break;
	 } else {
            /*
             * There are not enough pages -- drop the locks and wait for 
             * the host and/or other VMs to produce free pages.
	     */ 
            HostIF_VMUnlock(vm, 0);
	    HostIF_GlobalUnlock(17);
	    HostIF_WaitForFreePages(10);
	 }
      }
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86UnreserveFreePages --
 *
 *       Decreases the global and VM's locked page counts. 
 *       The function is thread-safe. 
 *
 * Results:
 *       void
 *
 * Side effects:
 *       The global lock and VM's lock are acquired and released.
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86UnreserveFreePages(VMDriver *vm, 
			unsigned int numPages)
{
   ASSERT(vm);

   HostIF_GlobalLock(18);
   HostIF_VMLock(vm, 1);

   ASSERT(numLockedPages >= numPages);
   ASSERT(vm->memInfo.locked >= numPages);

   numLockedPages -= numPages;
   vm->memInfo.locked -= numPages;
   vm->stats.monitorLockedPages -= numPages;

   HostIF_VMUnlock(vm, 1);
   HostIF_GlobalUnlock(18);
}


/*
 *-----------------------------------------------------------------------------
 *
 * Vmx86_CreateVM --
 *
 *      Allocate and initialize a driver structure for a virtual machine.
 *      As a complimentary duty, check if dedicated zero COW page exists 
 *      and, if it does not, allocate one.
 *
 * Results:
 *      VMDriver structure or NULL on error.
 *
 * Side effects:
 *      None
 *
 *-----------------------------------------------------------------------------
 */

VMDriver *
Vmx86_CreateVM(void *uniqHandle,             // IN: Unique id for VM to be
                                             //     created
	       uintptr_t processId,          // IN: Process creating VM
	       enum vmEmulations vmVersion)
{
   VMDriver *vm;
   Vcpuid v;

   HostIF_GlobalLock(20);
   if (cowZeroPage == NULL) {
      /*
       * This can only happen if at vmmon's upload time allocation
       * of a dedicated zero page failed. If so, let us try again
       * at power on time for a VM. Failing to allocate a page at
       * this point is a fair reason not to create a VM.
       */
      uint64 cowZeroKey;
      memset(cowBuf, 0, PAGE_SIZE);
      cowZeroKey = Hash_Page(cowBuf);
      cowZeroPage = Vmx86COWAllocPage(NULL, cowZeroKey, cowBuf, TRUE);
      if (cowZeroPage == NULL) {
         HostIF_GlobalUnlock(20);
         return NULL;
      }
   }
   HostIF_GlobalUnlock(20);

   vm = HostIF_AllocKernelMem(sizeof *vm, TRUE);
   if (vm == NULL) {
      return NULL;
   }
   memset(vm, 0, sizeof *vm);

   ASSERT(sizeof (uintptr_t) == sizeof vm);
   vm->id = (unsigned)(uintptr_t)vm >> 1;
   vm->uniqHandle = uniqHandle;
   vm->processID = processId;
   vm->memInfo.admitted = FALSE;
   vm->hintIndex = 0;
   vm->cowID = INVALID_COWID;
   vm->fastMonSwitch.performingSwitch = FALSE;
   vm->fastMonSwitch.oldVmId = -1;
   for (v = 0; v < MAX_INITBLOCK_CPUS; v++) {
      vm->currentHostCpu[v] = INVALID_HOST_CPU;
   }

   if (HostIF_Init(vm)) {
      goto cleanup;
   }

   HostIF_GlobalLock(0);

   if (vmCount >= MAX_VMS) {
      HostIF_GlobalUnlock(0);
      goto cleanup;
   }

   switch (vmVersion) {
      case VME_TOT:
      case VME_V6:
      case VME_S1B1:
      case VME_V55:
      case VME_V5:
      case VME_GSX32:
      case VME_V452:
      case VME_V45:
      case VME_V4:
      case VME_GSX251:
      case VME_GSX25:
      case VME_GSX2:
      case VME_V321:
      case VME_V32:
      case VME_V3:
      case VME_V2:
      case VME_GSX1:
	 break;
      default:
         HostIF_GlobalUnlock(0);
	 printk(KERN_ERR "Unknown emulation %u\n", vmVersion);
	 goto cleanup;
   }
   vm->vmVersion = vmVersion;
   Vmx86RegisterVMOnList(vm);

   HostIF_GlobalUnlock(0);

   return vm;

cleanup:
   /* 
    * The VM is not on a list, "vmCount" has not been incremented,
    * "vm->cowID" is INVALID_COWID, and either the VM's mutex hasn't
    * been initialized or we've only taken the global lock and checked
    * a counter since, so we know that the VM has not yet locked any
    * pages.
    */
   ASSERT(vm->memInfo.locked == 0);
   Vmx86FreeAllVMResources(vm);
   return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ReleaseVM  --
 *
 *      Release a VM (either created here or from a bind).
 *
 * Results:
 *      zero if successful
 *
 * Side effects:
 *	Decrement VM reference count.
 *      Release resources (those that are left) when count reaches 0.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_ReleaseVM(VMDriver *vm)
{
   HostIF_GlobalLock(1);

   /*
    * Do reference counting first
    */

   if (--vm->count > 0) {
      HostIF_GlobalUnlock(1);
      return 0;
   }

   ASSERT(vm->count == 0);

   Vmx86DeleteVMFromList(vm);
   HostIF_GlobalUnlock(1);
   Vmx86FreeAllVMResources(vm);

   return 0;
}


/*
 *------------------------------------------------------------------------------
 *
 * Vmx86_InitNUMAInfo --
 *
 *    Initializaiton of the NUMA structures in the vmmon.  
 *    Expects all initial arguments to be part of the NUMAInfoArgs structure.
 *
 * Results:
 *    TRUE on success
 *    FALSE on failure
 *
 * Side effects:
 *    None.
 *
 *------------------------------------------------------------------------------
 */

Bool
Vmx86_InitNUMAInfo(NUMAInfoArgs *initParams) // IN 
{
   unsigned int i, j, cnt = 0;
   if (initParams == NULL) {
      return FALSE;
   }
   if (initParams->numNodes <= 0 || initParams->numMemRanges <= 0 ||
       initParams->numNodes > NUMA_MAX_NODES ||
       initParams->numMemRanges > (initParams->numNodes * NUMA_MAX_MEM_RANGES)) {
      return FALSE;
   }

   /* Locking for the NUMA structres, which will be shared across all VMs*/
   HostIF_GlobalLock(27);

   /* If its greater than 0, the structures have already been initialized*/
   if (numaNumNodes > 0) {
      goto succeed;
   }
   
   numaNodes = HostIF_AllocKernelMem((sizeof(NUMA_NodeInfo) * 
                                      initParams->numNodes), TRUE);
   apicToNUMANode = HostIF_AllocKernelMem((sizeof(*apicToNUMANode) * 
                                           initParams->numNodes * 
                                           NUMA_MAX_CPUS_PER_NODE), TRUE);
   numaMemRangesList = HostIF_AllocKernelMem((sizeof(NUMA_MemRange) * 
                                              initParams->numMemRanges),TRUE);

   if (!numaNodes || !apicToNUMANode || !numaMemRangesList) {
      goto abort;
   } 

   /*
    *  Should not memcpy NUMA_NodeInfo * NUMA_MAX_NODES, since we cut down
    *  the buffer to the number of numaNodes that we actually pass down.
    */
   memcpy(numaNodes, initParams->nodes, 
          sizeof(NUMA_NodeInfo) * (initParams->numNodes));
   
   numaNumNodes = initParams->numNodes;
   numaNumMemRanges = initParams->numMemRanges;
   
   /*
    *  Updating the numaMemRangesList and apicTonode structures so its faster
    *  for lookups
    */
   for (i = 0; i < numaNumNodes; i++) {
      for (j = 0; j < numaNodes[i].numMemRanges; j++) {
         ASSERT(cnt < numaNumMemRanges);
         memcpy(&numaMemRangesList[cnt], &numaNodes[i].memRange[j], 
                sizeof(NUMA_MemRange));
	 cnt++;
      }
      for (j = 0; j < numaNodes[i].numPCPUs; j++) {
         ASSERT(numaNodes[i].apicIDs[j] < (numaNumNodes * NUMA_MAX_CPUS_PER_NODE));
         apicToNUMANode[numaNodes[i].apicIDs[j]] = numaNodes[i].id; 
      }
 
   }
   ASSERT(cnt == numaNumMemRanges);
   ASSERT(cnt < NUMA_MAX_TOTAL_MEM_RANGES);
   Log("Vmx86_InitNUMAInfo : numaNumMemRanges=%d and numaNumNodes=%d\n",
       numaNumMemRanges, numaNumNodes);

succeed:
   HostIF_GlobalUnlock(27);
   return TRUE;

abort:
   Vmx86_DestroyNUMAInfo(); 
   HostIF_GlobalUnlock(27);
   return FALSE;   
}


/*
 *------------------------------------------------------------------------------
 *
 * Vmx86_DestroyNUMAInfo --
 *
 *    Destruction of the NUMA structures in the vmmon.  
 *    Will happen during unloading of the vmmon.
 *
 * Results:
 *    TRUE on success
 *
 * Side effects:
 *    None.
 *
 *------------------------------------------------------------------------------
 */

void
Vmx86_DestroyNUMAInfo(void) 
{
   if (numaNodes) {
      HostIF_FreeKernelMem(numaNodes);
   }
   if (apicToNUMANode) {
      HostIF_FreeKernelMem(apicToNUMANode);
   }
   if (numaMemRangesList) {
      HostIF_FreeKernelMem(numaMemRangesList);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * NUMA_MPNToNodeNum --
 *
 *      Returns the node corresponding to the given machine page.
 *
 * Results:
 *      node or INVALID_NUMANODE if the MPN is not within any
 *      node's memory ranges
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
*/

NUMA_Node
Vmx86_MPNToNodeNum(MPN mpn) //IN
{
   unsigned int i;

   for (i = 0; i < numaNumMemRanges; i++) {
      if (mpn >= numaMemRangesList[i].startMPN &&
          mpn <= numaMemRangesList[i].endMPN ) {
         return numaMemRangesList[i].id ;
      }	
   }

   return INVALID_NUMANODE;
}


/*
 *------------------------------------------------------------------------------
 *
 * Vmx86_GetNUMAMemStats --
 *
 *    Gets statistics about the current VM memory usage from the static 
 *    NUMA structures that were initialized by the VMx86_InitNUMAInfo function.
 *    Packs all the results in the VMNUMAMemStatsArgs structure and returns.
 *
 * Results:
 *    TRUE on success
 *    FALSE on failure
 *
 * Side effects:
 *    None.
 *
 *------------------------------------------------------------------------------
 */

Bool
Vmx86_GetNUMAMemStats(VMDriver *curVM,            //IN
                      VMNUMAMemStatsArgs *outArgs)//OUT: stats for curVM
{
   VMDriver *vm;
   int i;
   NUMA_Node curNode;
   int apicId;
   DEBUG_ONLY(int initialApicId;)
   
   ASSERT(curVM);
   
   apicId = HostIF_APIC_ID();
   
   if (apicId == APIC_INVALID_ID) {
      Warning("Vmx86_GetNUMAMemStats: cant read LAPIC ID\n");
      return FALSE;
   }
   curNode = apicToNUMANode[apicId];

   if (curNode >= NUMA_MAX_NODES) {
      Warning("Vmx86_GetNUMAMemStats: Invalid node\n");
      return FALSE;
   }

   outArgs->curNUMANode = curNode;

   for (i = 0; i < numaNodes[curNode].numPCPUs; i++) {
      if (numaNodes[curNode].apicIDs[i] == apicId ) {
         outArgs->curCPU = i;
      }
   }
   if (i == NUMA_MAX_CPUS_PER_NODE) {
      Warning("Processor not part of this node, structures are wrong \n");
   }
   
   /*
    * The inital APIC is read from the CPUID instruction, whereas the current
    * is read from a physical address, done by the HostIF_APIC_ID().
    * Confirm that they are same.
    */   

   DEBUG_ONLY({
      initialApicId = HostIF_GetInitialApicId();
      if (apicId != initialApicId) {
         Warning("Vmx86_GetNUMAStats: The current APIC Id of the processor is "
                 "different from the Inital value\n");
      }
   })

   vm = curVM;
   HostIF_VMLock(vm, 17);

   /* 
    * We dont have access to the VMHost structure members in this file, 
    * but we do have it in the hostif.c file. Hence, this function simply,
    * does the scanning of anonymous pages and gives the result back.
    */
   ASSERT(sizeof outArgs->numPagesPerNode == (NUMA_MAX_NODES * sizeof(uint32)));
   if (!HostIF_GetNUMAAnonPageDistribution(vm, NUMA_MAX_NODES, 
                                           outArgs->numPagesPerNode)) {
      Log("VM has no anonymous pages\n");
   }
   HostIF_VMUnlock(vm, 17);
  
   return TRUE;
}


/*
 *------------------------------------------------------------------------------
 *
 * Vmx86_InitVM --
 *
 *    Initializaiton of the VM.  Expects all initial arguments
 *    to be part of the InitBlock structure.
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    Many
 *
 *------------------------------------------------------------------------------
 */

int
Vmx86_InitVM(VMDriver *vm,          // IN
             InitBlock *initParams) // IN: Initial params from the VM
{
   int retval;

   if (initParams->magicNumber != INIT_BLOCK_MAGIC) {
      Warning("Bad magic number for init block 0x%x\n", initParams->magicNumber);
      return 1;
   }
   if (initParams->numVCPUs >= MAX_INITBLOCK_CPUS) {
      Warning("Too many VCPUs for init block %d\n", initParams->numVCPUs);
      return 1;
   }
   vm->numVCPUs = initParams->numVCPUs;

   HostIF_InitFP(vm);
   HostIF_InitEvent(vm);

   /*
    * Initialize the driver's part of the cross-over page used to
    * talk to the monitor
    */

   retval = Task_InitCrosspage(vm, initParams);
   if (retval) {
      Warning("Task crosspage init died with retval=%d\n", retval);
      /*
       *  Note that any clean-up of resources will be handled during
       *  power-off when Vmx86_ReleaseVM() is called as part of
       *  MonitorLoop_PowerOff(). 
       */
      return 1;
   }   

   /*
    *  Check if we want to arbitrarily fail every N VM initializations.
    *  Useful in testing PR 72482.
    */
   if (initParams->vmInitFailurePeriod != 0) {
      static uint32 counter = 0;
      if ((++counter) % initParams->vmInitFailurePeriod == 0) {
         Warning("VM initialization failed on %d iteration\n", counter);
         return 1;
      }
   }

   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_LateInitVM --
 *
 *      Do late initialization of the driver.
 *	This should be called after Vmx86_CreateVM and
 *	after all the user-level device initialization.
 *
 * Results: 
 *	non-zero on error, zero on success;
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_LateInitVM(VMDriver *vm)
{
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetStartTime --
 *
 *      Initial the data structures that track the starting time of the
 *      virtual machine.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      none
 *
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetStartTime(VmTimeStart *st)	// OUT: return value
{
   uintptr_t flags;

   SAVE_FLAGS(flags);
   CLEAR_INTERRUPTS();

   st->count = RDTSC();
   st->time = HostIF_ReadUptime();

   RESTORE_FLAGS(flags);
}


#ifdef __APPLE__
/*
 *----------------------------------------------------------------------
 *
 * Vmx86GetBusyKHzEstimate
 *
 *      Return an estimate the of the processor's kHz rating, based on
 *      a spinloop.  This is especially useful on systems where the TSC
 *      is known to run at its maximum rate when we are using the CPU.
 *      As of 2006, Intel Macs are this way... the TSC rate is 0 if the
 *      CPU is in a deep enough sleep state, or at its max rate otherwise.
 *
 * Results:
 *      Processor speed in kHz.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER uint32
Vmx86GetBusyKHzEstimate(void)
{
   static const int ITERS = 100;
   static const int CYCLES_PER_ITER = 20000;
   int i;
   uint64 j;
   uint64 freq;
   uint64 hz;
   uint64 aggregateCycles = 0;
   uint64 aggregateUptime = 0;

   for (i = 0; i < ITERS; i++) {
      NO_INTERRUPTS_BEGIN() {
         aggregateCycles -= RDTSC();
         aggregateUptime -= HostIF_ReadUptime();
         for (j = RDTSC() + CYCLES_PER_ITER; RDTSC() < j; )
            ;
         aggregateCycles += RDTSC();
         aggregateUptime += HostIF_ReadUptime();
      } NO_INTERRUPTS_END();
   }
   freq = HostIF_UptimeFrequency();
   while (aggregateCycles > MAX_UINT64 / freq) {
      aggregateCycles >>= 1;
      aggregateUptime >>= 1;
   }
   hz = aggregateCycles * freq / aggregateUptime;
   return (hz + 500) / 1000;
}
#endif


/*
 *----------------------------------------------------------------------
 *
 * Vmx86GetkHzEstimate
 *
 *      Return an estimate of the processor's kHz rating, based on
 *      the ratio of the cycle counter and system uptime since the
 *      driver was loaded.
 *      This function could be called (on Windows) at IRQL DISPATCH_LEVEL.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER uint32
Vmx86GetkHzEstimate(VmTimeStart *st)	// IN: start time
{
   uint64 cDiff, tDiff, freq, hz;
   uintptr_t flags;
   uint32 kHz = 0;

   SAVE_FLAGS(flags);
   CLEAR_INTERRUPTS();
   cDiff = RDTSC() - st->count;
   tDiff = HostIF_ReadUptime() - st->time;
   RESTORE_FLAGS(flags);

   if (tDiff == 0) {
      goto failure;
   }

   /*
    * Compute the CPU speed in kHz, which is cDiff / (tDiff /
    * HostIF_UptimeFrequency()) / 1000.  We need to do the computation
    * carefully to avoid overflow or undue loss of precision.  Also,
    * on Linux we can't do a 64/64=64 bit division directly, as the
    * gcc stub for that is not linked into the kernel.
    */
   freq = HostIF_UptimeFrequency();
#if defined VM_X86_64 || !defined linux
   while (cDiff > ((uint64) -1) / freq) {
      cDiff >>= 1;
      tDiff >>= 1;
   }
   hz  = (cDiff * freq) / tDiff;
   kHz = (uint32) ((hz + 500) / 1000);
#else
   {
      uint32 tmpkHz, tmp;
      /* On Linux we can't do a 64/64=64 bit division, as the gcc stub
       * for that is not linked into the kernel.  We'll assume that cDiff
       * * freq fits into 64 bits and that tDiff fits into 32 bits.  This
       * is safe given the values used on Linux.
       */
      Div643264(cDiff * freq, tDiff, &hz, &tmp);
      hz += 500;
      /*
       * If result in kHz cannot fit into 32 bits, we would get a divide
       * by zero exception.
       */
      if ((uint32)(hz >> 32) >= 1000) {
         goto failure;
      }
      Div643232(hz, 1000, &tmpkHz, &tmp);
      kHz = tmpkHz;
   }
#endif
   return kHz;

failure:
#ifdef VMW_HAS_CPU_KHZ
   /* If we have some reasonable value, use it... */
   kHz = cpu_khz;
#endif
   return kHz;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetkHzEstimate
 *
 *      Return an estimate of the processor's kHz rating, based on
 *      the ratio of the cycle counter and system uptime since the
 *      driver was loaded.  Or based on a spinloop.
 *
 *      This function could be called (on Windows) at IRQL DISPATCH_LEVEL.
 *
 * Results:
 *      Processor speed in kHz.
 *
 * Side effects:
 *      Result is cached.
 *
 *----------------------------------------------------------------------
 */

uint32
Vmx86_GetkHzEstimate(VmTimeStart *st)	// IN: start time
{
   static uint32 kHz; 

   /* 
    * Cache and return the first result for consistency. 
    * TSC values can be changed without notification.
    * TSC frequency can be vary too (SpeedStep, slowing clock on HALT, etc.)
    */
   if (kHz != 0) {
      return kHz;
   }

#ifdef __APPLE__
   return kHz = Vmx86GetBusyKHzEstimate();
#else
   return kHz = Vmx86GetkHzEstimate(st);
#endif
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetHostClockRate --
 *
 *      The monitor wants to poll for events at the given rate. If no VM
 *      is specified, then 'rate' is ignored and the last set rate is set
 *      again.
 *
 * Results:
 *      0 for success, host-specific error code for failure.
 *
 * Side effects:
 *      May increase the host timer interrupt rate, etc.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_SetHostClockRate(VMDriver *vm,  // IN: VM instance pointer
                       int rate)      // IN: rate in Hz
{
   unsigned newGlobalRate;
   VMDriver *cur;
   int retval = 0;

   if (!vm) {
      Log("Resetting last set host clock rate of %d\n", globalFastClockRate);
      HostIF_FastClockLock(0);
      retval = HostIF_SetFastClockRate(globalFastClockRate);
      HostIF_FastClockUnlock(0);
      return retval;
   }

   /* Quick test before locks are acquired. */
   if (vm->fastClockRate == rate) {
      return retval;
   }

   HostIF_FastClockLock(2);
   if (vm->fastClockRate == rate) {
      HostIF_FastClockUnlock(2);
      return retval;
   }

   /*
    * Loop through all vms to find new max rate.
    */
   newGlobalRate = rate;
   HostIF_GlobalLock(19);
   for (cur = vmDriverList; cur != NULL; cur = cur->nextDriver) {
      if (cur != vm && cur->fastClockRate > newGlobalRate) {
         newGlobalRate = cur->fastClockRate;
      }
   }
   HostIF_GlobalUnlock(19);

   if (newGlobalRate != globalFastClockRate) {
      Log("host clock rate change request %d -> %d\n",
          globalFastClockRate, newGlobalRate);
      retval = HostIF_SetFastClockRate(newGlobalRate);
      if (!retval) {
         globalFastClockRate = newGlobalRate;
      }
   }
   if (!retval) {
      vm->fastClockRate = rate;
   }
   HostIF_FastClockUnlock(2);
   
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_MonitorPollIPI --
 *
 *      Check for VCPUs that are in the monitor and need an IPI to
 *      fire their next MonitorPoll callback.  Should be called once
 *      per fast timer interrupt if the fast timer is in use.
 *      Otherwise does not need to be called at all, as the normal
 *      timer interrupts will wake up MonitorPoll often enough.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      May send IPIs.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER int Vmx86IPIV45(VMDriver *vm,
					    VmAbsoluteTS rNow)
{
   VmAbsoluteTS pNow;
   Vcpuid v, vv;
   VCPUSet firstVCS;
   VmAbsoluteTS t, firstTS;
   VMCrossPageV45* crosspage;

   crosspage = vm->crosspage[0];
   if (!crosspage) {
      return 0;  // VCPU is not initialized yet
   }
   /*
    * Convert real TSC to pseudo TSC.  This code will have to
    * change when/if we start to support MP hosts with
    * unsynchronized TSCs (see PR 20499), but for now it's OK to
    * just use the tscAdjustment from VCPU 0's crosspage.
    */
   pNow = rNow + crosspage->tscAdjustment;

   /*
    * Loop through VCPUs in this VM to find the most up-to-date
    * copy of <monitorPollFirstTS, monitorPollFirstVCPU>.
    */
   vv = 0;
   firstVCS = crosspage->monitorPollFirstVCS;
   firstTS = crosspage->monitorPollFirstTS;
   for (v = 1; v < vm->numVCPUs; v++) {
      crosspage = vm->crosspage[v];
      if (!crosspage) {
         continue;  // VCPU is not initialized yet
      }
      t = crosspage->monitorPollFirstTS;
      if (t && COMPARE_TS(t, >, firstTS)) {
         firstVCS = crosspage->monitorPollFirstVCS;
         firstTS = t;
         vv = v;  // remember where we found it
      }
   }
         
   /*
    * Check if it's time to send an IPI
    */
   if (!VCPUSet_IsEmpty(firstVCS) && COMPARE_TS(pNow, >, firstTS)) {
      Bool sentIPI = HostIF_IPI(vm, firstVCS, FALSE);
      if (sentIPI) {
         ((typeof(crosspage))vm->crosspage[vv])->monitorPollFirstVCS = VCPUSet_Empty(); // done
#if defined(linux)
         // On Linux, HostIF_IPI is always a broadcast, so there is no
         // point in doing it for more than one VM.
         return 1;
#endif
      }
   }
   return 0;
}

/* Identical to Vmx86IPIV45, except crosspage type & function name */
static INLINE_SINGLE_CALLER int Vmx86IPIV5(VMDriver *vm,
					     VmAbsoluteTS rNow)
{
   VmAbsoluteTS pNow;
   Vcpuid v, vv;
   VCPUSet firstVCS;
   VmAbsoluteTS t, firstTS;
   VMCrossPageV5* crosspage;

   crosspage = vm->crosspage[0];
   if (!crosspage) {
      return 0;  // VCPU is not initialized yet
   }
   /*
    * Convert real TSC to pseudo TSC.  This code will have to
    * change when/if we start to support MP hosts with
    * unsynchronized TSCs (see PR 20499), but for now it's OK to
    * just use the tscAdjustment from VCPU 0's crosspage.
    */
   pNow = rNow + crosspage->tscAdjustment;

   /*
    * Loop through VCPUs in this VM to find the most up-to-date
    * copy of <monitorPollFirstTS, monitorPollFirstVCPU>.
    */
   vv = 0;
   firstVCS = crosspage->monitorPollFirstVCS;
   firstTS = crosspage->monitorPollFirstTS;
   for (v = 1; v < vm->numVCPUs; v++) {
      crosspage = vm->crosspage[v];
      if (!crosspage) {
         continue;  // VCPU is not initialized yet
      }
      t = crosspage->monitorPollFirstTS;
      if (t && COMPARE_TS(t, >, firstTS)) {
         firstVCS = crosspage->monitorPollFirstVCS;
         firstTS = t;
         vv = v;  // remember where we found it
      }
   }
         
   /*
    * Check if it's time to send an IPI
    */
   if (!VCPUSet_IsEmpty(firstVCS) && COMPARE_TS(pNow, >, firstTS)) {
      Bool sentIPI = HostIF_IPI(vm, firstVCS, FALSE);
      if (sentIPI) {
         ((typeof(crosspage))vm->crosspage[vv])->monitorPollFirstVCS = VCPUSet_Empty(); // done
#if defined(linux)
         // On Linux, HostIF_IPI is always a broadcast, so there is no
         // point in doing it for more than one VM.
         return 1;
#endif
      }
   }
   return 0;
}

/* Identical to Vmx86IPIV45, except crosspage type & function name */
static INLINE_SINGLE_CALLER int Vmx86IPIV55(VMDriver *vm,
					    VmAbsoluteTS rNow)
{
   VmAbsoluteTS pNow;
   Vcpuid v, vv;
   VCPUSet firstVCS;
   VmAbsoluteTS t, firstTS;
   VMCrossPageV55* crosspage;

   crosspage = vm->crosspage[0];
   if (!crosspage) {
      return 0;  // VCPU is not initialized yet
   }
   /*
    * Convert real TSC to pseudo TSC.  This code will have to
    * change when/if we start to support MP hosts with
    * unsynchronized TSCs (see PR 20499), but for now it's OK to
    * just use the tscAdjustment from VCPU 0's crosspage.
    */
   pNow = rNow + crosspage->tscAdjustment;

   /*
    * Loop through VCPUs in this VM to find the most up-to-date
    * copy of <monitorPollFirstTS, monitorPollFirstVCPU>.
    */
   vv = 0;
   firstVCS = crosspage->monitorPollFirstVCS;
   firstTS = crosspage->monitorPollFirstTS;
   for (v = 1; v < vm->numVCPUs; v++) {
      crosspage = vm->crosspage[v];
      if (!crosspage) {
         continue;  // VCPU is not initialized yet
      }
      t = crosspage->monitorPollFirstTS;
      if (t && COMPARE_TS(t, >, firstTS)) {
         firstVCS = crosspage->monitorPollFirstVCS;
         firstTS = t;
         vv = v;  // remember where we found it
      }
   }
         
   /*
    * Check if it's time to send an IPI
    */
   if (!VCPUSet_IsEmpty(firstVCS) && COMPARE_TS(pNow, >, firstTS)) {
      Bool sentIPI = HostIF_IPI(vm, firstVCS, FALSE);
      if (sentIPI) {
         ((typeof(crosspage))vm->crosspage[vv])->monitorPollFirstVCS = VCPUSet_Empty(); // done
#if defined(linux)
         // On Linux, HostIF_IPI is always a broadcast, so there is no
         // point in doing it for more than one VM.
         return 1;
#endif
      }
   }
   return 0;
}

/* Identical to Vmx86IPIV45, except crosspage type & function name */
static INLINE_SINGLE_CALLER int Vmx86IPIS1B1(VMDriver *vm,
					     VmAbsoluteTS rNow)
{
   VmAbsoluteTS pNow;
   Vcpuid v, vv;
   VCPUSet firstVCS;
   VmAbsoluteTS t, firstTS;
   VMCrossPageS1B1* crosspage;

   crosspage = vm->crosspage[0];
   if (!crosspage) {
      return 0;  // VCPU is not initialized yet
   }
   /*
    * Convert real TSC to pseudo TSC.  This code will have to
    * change when/if we start to support MP hosts with
    * unsynchronized TSCs (see PR 20499), but for now it's OK to
    * just use the tscAdjustment from VCPU 0's crosspage.
    */
   pNow = rNow + crosspage->tscAdjustment;

   /*
    * Loop through VCPUs in this VM to find the most up-to-date
    * copy of <monitorPollFirstTS, monitorPollFirstVCPU>.
    */
   vv = 0;
   firstVCS = crosspage->monitorPollFirstVCS;
   firstTS = crosspage->monitorPollFirstTS;
   for (v = 1; v < vm->numVCPUs; v++) {
      crosspage = vm->crosspage[v];
      if (!crosspage) {
         continue;  // VCPU is not initialized yet
      }
      t = crosspage->monitorPollFirstTS;
      if (t && COMPARE_TS(t, >, firstTS)) {
         firstVCS = crosspage->monitorPollFirstVCS;
         firstTS = t;
         vv = v;  // remember where we found it
      }
   }
         
   /*
    * Check if it's time to send an IPI
    */
   if (!VCPUSet_IsEmpty(firstVCS) && COMPARE_TS(pNow, >, firstTS)) {
      Bool sentIPI = HostIF_IPI(vm, firstVCS, FALSE);
      if (sentIPI) {
         ((typeof(crosspage))vm->crosspage[vv])->monitorPollFirstVCS = VCPUSet_Empty(); // done
#if defined(linux)
         // On Linux, HostIF_IPI is always a broadcast, so there is no
         // point in doing it for more than one VM.
         return 1;
#endif
      }
   }
   return 0;
}

static INLINE_SINGLE_CALLER int Vmx86IPIV6(VMDriver *vm,
					   VmAbsoluteTS rNow)
{
      VmAbsoluteTS pNow;
      Vcpuid v, vv;
      VCPUSet firstVCS;
      VmAbsoluteTS t, firstTS;
      uintptr_t flags;
      VMCrossPageV6 *crosspage;

      /*
       * Convert real TSC to pseudo TSC, using the pseudoTSCConv from
       * VCPU 0's crosspage. This is accurate enough since it will
       * usually been updated recently, especially in the case we care
       * about where some VCPU of this VM is currently in the monitor.
       * In the future when/if we start to support MP hosts with
       * unsynchronized TSCs (see PR 20499), we'll need to use a
       * pseudoTSCConv that is correct for the current PCPU, but at
       * present they are not PCPU-specific.
       */
      SAVE_FLAGS(flags);
      CLEAR_INTERRUPTS();
      crosspage = vm->crosspage[0];
      if (!crosspage) {
         RESTORE_FLAGS(flags);
         return 0;  // VCPU is not initialized yet
      }
      pNow = RateConv_Unsigned(&crosspage->pseudoTSCConv, rNow);
      RESTORE_FLAGS(flags);

      /*
       * Loop through VCPUs in this VM to find the most up-to-date
       * copy of <monitorPollFirstTS, monitorPollFirstVCPU>.
       */
      vv = 0;
      firstVCS = crosspage->monitorPollFirstVCS;
      firstTS = crosspage->monitorPollFirstTS;
      for (v = 1; v < vm->numVCPUs; v++) {
         crosspage = vm->crosspage[v];
         if (!crosspage) {
            continue;  // VCPU is not initialized yet
         }
         t = crosspage->monitorPollFirstTS;
         if (t && COMPARE_TS(t, >, firstTS)) {
            firstVCS = crosspage->monitorPollFirstVCS;
            firstTS = t;
            vv = v;  // remember where we found it
         }
      }
         
      /*
       * Check if it's time to send an IPI
       */
      if (!VCPUSet_IsEmpty(firstVCS) && COMPARE_TS(pNow, >, firstTS)) {
         Bool sentIPI = HostIF_IPI(vm, firstVCS, FALSE);
         if (sentIPI) {
            ((typeof(crosspage))vm->crosspage[vv])->monitorPollFirstVCS = VCPUSet_Empty(); // done
#if defined(linux)
            // On Linux, HostIF_IPI is always a broadcast, so there is no
            // point in doing it for more than one VM.
            return 1;
#endif
         }
      }
      return 0;
}

static INLINE_SINGLE_CALLER int Vmx86IPITOT(VMDriver *vm,
					    VmAbsoluteTS rNow)
{
      VmAbsoluteTS pNow;
      Vcpuid v, vv;
      VCPUSet firstVCS;
      VmAbsoluteTS t, firstTS;
      uintptr_t flags;
      VMCrossPageTOT *crosspage;

      /*
       * Convert real TSC to pseudo TSC, using the pseudoTSCConv from
       * VCPU 0's crosspage. This is accurate enough since it will
       * usually been updated recently, especially in the case we care
       * about where some VCPU of this VM is currently in the monitor.
       * In the future when/if we start to support MP hosts with
       * unsynchronized TSCs (see PR 20499), we'll need to use a
       * pseudoTSCConv that is correct for the current PCPU, but at
       * present they are not PCPU-specific.
       */
      SAVE_FLAGS(flags);
      CLEAR_INTERRUPTS();
      crosspage = vm->crosspage[0];
      if (!crosspage) {
         RESTORE_FLAGS(flags);
         return 0;  // VCPU is not initialized yet
      }
      pNow = RateConv_Unsigned(&crosspage->pseudoTSCConv, rNow);
      RESTORE_FLAGS(flags);

      /*
       * Loop through VCPUs in this VM to find the most up-to-date
       * copy of <monitorPollFirstTS, monitorPollFirstVCPU>.
       */
      vv = 0;
      firstVCS = crosspage->monitorPollFirstVCS;
      firstTS = crosspage->monitorPollFirstTS;
      for (v = 1; v < vm->numVCPUs; v++) {
         crosspage = vm->crosspage[v];
         if (!crosspage) {
            continue;  // VCPU is not initialized yet
         }
         t = crosspage->monitorPollFirstTS;
         if (t && COMPARE_TS(t, >, firstTS)) {
            firstVCS = crosspage->monitorPollFirstVCS;
            firstTS = t;
            vv = v;  // remember where we found it
         }
      }
         
      /*
       * Check if it's time to send an IPI
       */
      if (!VCPUSet_IsEmpty(firstVCS) && COMPARE_TS(pNow, >, firstTS)) {
         Bool sentIPI = HostIF_IPI(vm, firstVCS, FALSE);
         if (sentIPI) {
            ((typeof(crosspage))vm->crosspage[vv])->monitorPollFirstVCS = VCPUSet_Empty(); // done
#if defined(linux)
            // On Linux, HostIF_IPI is always a broadcast, so there is no
            // point in doing it for more than one VM.
            return 1;
#endif
         }
      }
      return 0;
}

void
Vmx86_MonitorPollIPI(void)
{
   VMDriver *vm;
   VmAbsoluteTS rNow = RDTSC();

   /*
    * Loop through all vms -- needs the global lock to protect vmDriverList.
    */
   HostIF_GlobalLock(21);
   for (vm = vmDriverList; vm != NULL; vm = vm->nextDriver) {
      int stop;

      switch (vm->vmVersion) {
         case VME_V45:  
	 case VME_V452:
	 case VME_GSX32:
	 		stop = Vmx86IPIV45(vm, rNow); break;
	 case VME_V5:   stop = Vmx86IPIV5(vm, rNow); break;
	 case VME_V55:  stop = Vmx86IPIV55(vm, rNow); break;
	 case VME_S1B1: stop = Vmx86IPIS1B1(vm, rNow); break;
	 case VME_V6:   stop = Vmx86IPIV6(vm, rNow); break;
	 case VME_TOT:  stop = Vmx86IPITOT(vm, rNow); break;
	 default:       stop = 0; break;
      }
      if (stop) {
         break;
      }
   }
   HostIF_GlobalUnlock(21);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetNumVMs  --
 *
 *      Return the number of VMs.
 *
 * Results:
 *      The number of VMs.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int32
Vmx86_GetNumVMs()
{
   return vmCount;
}

int32
Vmx86_GetTotalMemUsage()
{
   VMDriver *vm;
   int totalmem = 0;

   HostIF_GlobalLock(15);
   vm = vmDriverList;

   for (vm = vmDriverList; vm != NULL; vm = vm->nextDriver) {
      /*
       * The VM lock is not strictly necessary as the vm will
       * stay on the list until we release the global lock and
       * because of order in which "admitted" and "mainMemSize"
       * are set when each VM is admitted.
       */
      if (vm->memInfo.admitted || vm->vmVersion < VME_V55) {
          totalmem += PAGES_2_MBYTES(ROUNDUP(vm->memInfo.mainMemSize,
                                             MBYTES_2_PAGES(1)));
      }
   }
   
   HostIF_GlobalUnlock(15);
   return totalmem;
}


static INLINE unsigned
Vmx86MinAllocationFunc(unsigned paged, unsigned nonpaged, unsigned swappable,
                       unsigned memPct)
{
   unsigned newMin = (memPct * swappable) / 100 + 
                     (paged - swappable) + nonpaged;
   ASSERT(swappable <= paged);
   return newMin;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86MinAllocation --
 *
 *      Computes the minimum number of pages that must be allocated to a
 *      specific vm.  The minAllocation for a vm is defined as 
 *      some percentage of guest memory plus 100% of nonpagable (overhead) 
 *      memory.  
 * 
 * Results:
 *	The minAllocation for this vm.  
 *	
 *
 * Side effects:
 *      Analyzes the vm info, requiring the vm lock.
 *
 *----------------------------------------------------------------------
 */

static INLINE unsigned
Vmx86MinAllocation(VMDriver *vm, unsigned memPct) {
   ASSERT(HostIF_VMLockIsHeld(vm));
   return Vmx86MinAllocationFunc(vm->memInfo.paged, vm->memInfo.nonpaged,
                                 vm->memInfo.mainMemSize, memPct);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86CalculateGlobalMinAllocation --
 *
 *      Computes the sum of minimum allocations of each vm assuming a given
 *      percentage of guest memory must fit within host RAM. 
 *      
 * Results:
 *	Number of pages that must fit within host ram for a given overcommit
 *      level.
 *	
 *
 * Side effects:
 *      None. The actual minAllocations of each vm are NOT updated during
 *      this computation.
 *
 *----------------------------------------------------------------------
 */

static unsigned
Vmx86CalculateGlobalMinAllocation(unsigned memPct)
{
   VMDriver *vm;
   unsigned minAllocation = 0;
   
   ASSERT(HostIF_GlobalLockIsHeld());
   /* Pages of other vms required to fit inside the hard limit. */
   for (vm = vmDriverList; vm; vm = vm->nextDriver) {  
      HostIF_VMLock(vm, 2);
      if (vm->memInfo.admitted) {
         minAllocation += Vmx86MinAllocation(vm, memPct);
      }
      HostIF_VMUnlock(vm, 2);
   }
   return minAllocation;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86UpdateMinAllocations --
 *
 *      Updates the minimum allocation for each vm based on the global
 *      overcommitment percentage. 
 * 
 * Results:
 *      minAllocations for vms are changed.
 *	
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
Vmx86UpdateMinAllocations(unsigned memPct)
{
   VMDriver *vm;
   ASSERT(HostIF_GlobalLockIsHeld());
   /* Pages of other vms required to fit inside the hard limit. */
   for (vm = vmDriverList; vm; vm = vm->nextDriver) {
      HostIF_VMLock(vm, 3);
      if (vm->memInfo.admitted) {
         vm->memInfo.minAllocation = Vmx86MinAllocation(vm, memPct);
      }
      HostIF_VMUnlock(vm, 3);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetConfiguredLockedPagesLimit --
 *
 *      Set the user defined limit on the number of pages that can
 *      be locked.  This limit can be raised at any time but not lowered.
 *      This avoids having a user lower the limit as vms are running and
 *      inadvertently cause the vms to crash because of memory starvation.
 *
 * Results:
 *      Returns TRUE on success and FALSE on failure to set the limit
 *
 * Side effects:
 *      Hard limit may be changed.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_SetConfiguredLockedPagesLimit(unsigned limit)
{
   Bool retval = FALSE;

   HostIF_GlobalLock(4);
   if (limit >= lockedPageLimit.configured || vmCount == 1) {
      lockedPageLimit.configured = limit;
      retval = TRUE;
   }
   HostIF_GlobalUnlock(4);

   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetDynamicLockedPageLimit --
 *
 *      Set the dynamic locked page limit.  This limit is determined by
 *      authd in response to host pressure.  It can be both raised and
 *      lowered at any time.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Hard limit may be changed.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetDynamicLockedPagesLimit(unsigned limit)
{
   HostIF_GlobalLock(11);
   lockedPageLimit.dynamic = limit;
   HostIF_GlobalUnlock(11);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_LockPage --
 *
 *      Lock a page.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Number of global and per-VM locked pages increased.
 *
 *----------------------------------------------------------------------
 */

MPN
Vmx86_LockPage(VMDriver *vm,		     // IN: VMDriver
	       void* addr,		     // IN: VA of the page to lock
	       Bool allowMultipleMPNsPerVA)  // IN: allow locking many pages with the same VA
{
   MPN mpn;

   /* Atomically check and reserve locked memory */
   if (!Vmx86ReserveFreePages(vm, 1)) {
      return PAGE_LOCK_LIMIT_EXCEEDED;
   }

   HostIF_VMLock(vm, 4);
   mpn = HostIF_LockPage(vm, addr, allowMultipleMPNsPerVA);
   HostIF_VMUnlock(vm, 4);

   if (!PAGE_LOCK_SUCCESS(mpn)) {
      Vmx86UnreserveFreePages(vm, 1);
   }

   return mpn;
}



/*
 *----------------------------------------------------------------------
 *
 * Vmx86_UnlockPage --
 *
 *      Unlock a page.
 *
 * Results:
 *      
 *
 * Side effects:
 *      Number of global and per-VM locked pages decreased.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_UnlockPage(VMDriver *vm, void *addr)
{
   int retval;
   
   HostIF_VMLock(vm, 5);
   retval = HostIF_UnlockPage(vm, addr);
   HostIF_VMUnlock(vm, 5);

   if (PAGE_LOCK_SUCCESS(retval)) {
      Vmx86UnreserveFreePages(vm, 1);
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_UnlockPageByMPN --
 *
 *      Unlock a page.
 *
 * Results:
 *      
 *
 * Side effects:
 *      Number of global and per-VM locked pages decreased.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_UnlockPageByMPN(VMDriver *vm, // IN: VMDriver
		      MPN mpn,	    // IN: the page to unlock
		      void* addr)   // IN: optional valid VA for this MPN
{
   int retval;

   HostIF_VMLock(vm, 6);
   retval = HostIF_UnlockPageByMPN(vm, mpn, addr);
   HostIF_VMUnlock(vm, 6);

   if (PAGE_LOCK_SUCCESS(retval)) {
      Vmx86UnreserveFreePages(vm, 1);
   }
   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Vmx86_AllocLockedPages --
 *
 *      Allocate physical locked pages from the kernel.
 *
 *      Initially the pages are not mapped to any user or kernel 
 *      address space.
 *
 * Results:
 *      Non-negative value on partial/full completion: actual number of
 *      allocated MPNs. MPNs of the allocated pages are copied to 
 *      the caller's buffer.
 *
 *	Negative system specific error code on error (NTSTATUS on Windows,
 *      etc.)
 *
 * Side effects:
 *      Number of global and per-VM locked pages is increased.
 *
 *-----------------------------------------------------------------------------
 */

int
Vmx86_AllocLockedPages(VMDriver *vm,	     // IN: VMDriver
		       MPN32 *mpns,	     // OUT: User VA of an array for
                                             //      allocated MPNs.
		       unsigned numPages,    // IN: number of pages to allocate
		       Bool kernelMPNBuffer) // IN: is the MPN buffer in kernel or user address space?
{
   int allocatedPages;

   if (!Vmx86ReserveFreePages(vm, numPages)) {
      // XXX What kind of system-specific error code is that? --hpreg
      return PAGE_LOCK_LIMIT_EXCEEDED;
   }
   
   HostIF_VMLock(vm, 7);
   allocatedPages = HostIF_AllocLockedPages(vm, mpns, numPages, kernelMPNBuffer); 
   HostIF_VMUnlock(vm, 7);

   if (allocatedPages < 0) {
      Vmx86UnreserveFreePages(vm, numPages);
   } else if (allocatedPages < numPages) {
      Vmx86UnreserveFreePages(vm, numPages - allocatedPages);
   }

   return allocatedPages;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_FreeLockedPages --
 *
 *      Frees physical locked pages from the kernel previosly allocated 
 *      by Vmx86_AllocLockedPages().
 *
 * Results:
 *	0 on success,
 *	non-0 system specific error code on error (NTSTATUS on Windows, etc.)
 *
 * Side effects:
 *      Number of global and per-VM locked pages is decreased.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_FreeLockedPages(VMDriver *vm,	    // IN: VM instance pointer
		      const MPN32* mpns,    // IN: user or kernel array of MPNs to free 
		      unsigned numPages,    // IN: number of pages to free
		      Bool kernelMPNBuffer) // IN: is the MPN buffer in kernel or user address space?
{
   int ret;

   HostIF_VMLock(vm, 8);
   ret = HostIF_FreeLockedPages(vm, mpns, numPages, kernelMPNBuffer);
   HostIF_VMUnlock(vm, 8);

   if (ret == 0) {
      Vmx86UnreserveFreePages(vm, numPages);
   }

   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_IsAnonPage --
 *
 *      Queries the driver to see if the mpn is an anonymous page.
 *
 * Results:
 *      True if mpn is an anonymous page, false otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_IsAnonPage(VMDriver *vm,       // IN: VM instance pointer
                 const MPN32 mpn)    // IN: MPN we are asking about
{
   Bool ret;

   HostIF_VMLock(vm, 16);
   ret = HostIF_IsAnonPage(vm, mpn);
   HostIF_VMUnlock(vm, 16);
   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetLockedPageList --
 *
 *      puts MPNs of pages that were allocated by HostIF_AllocLockedPages()
 *      into user mode buffer.
 *
 * Results:
 *	non-negative number of the MPNs in the buffer on success.
 *	negative error code on error.
 *
 * Side effects:
 *      none
 *
 *----------------------------------------------------------------------
 */

int 
Vmx86_GetLockedPageList(VMDriver* vm,          // IN: VM instance pointer
                        MPN32* mpns,           // OUT: user mode buffer for MPNs
		        unsigned int numPages) // IN: size of the buffer in MPNs 
{
   int ret;
      
   HostIF_VMLock(vm, 9);
   ret = HostIF_GetLockedPageList(vm, mpns, numPages);
   HostIF_VMUnlock(vm, 9);
   
   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetMemInfo --
 *
 *      Return the info about all VMs.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      VMGetMemInfoArgs is filled in.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_GetMemInfo(VMDriver *curVM,
                 int32 curVMOnly,
                 VMMemInfoArgs *outArgs,
                 int outArgsLength)
{
   VMDriver *vm;
   int i;
   int outSize;
   int wantedVMs;

   HostIF_GlobalLock(7);

   if (curVMOnly) {
      wantedVMs = 1;
   } else {
      wantedVMs = vmCount;
   }

   outSize = VM_GET_MEM_INFO_SIZE(wantedVMs);
   if (outSize > outArgsLength) {
      HostIF_GlobalUnlock(7);
      return FALSE;
   }
   
   outArgs->numVMs = wantedVMs;
   outArgs->numLockedPages = numLockedPages;
   outArgs->maxLockedPages = Vmx86LockedPageLimit(curVM);
   outArgs->lockedPageLimit = lockedPageLimit;
   outArgs->globalMinAllocation = Vmx86CalculateGlobalMinAllocation(minVmMemPct);
   outArgs->minVmMemPct = minVmMemPct;
   outArgs->callerIndex = -1;
   Vmx86COWStats(&outArgs->cowInfo);

   if (curVM != NULL) {
      if (wantedVMs == 1) {
         outArgs->memInfo[0] = curVM->memInfo;
         outArgs->callerIndex = 0;
      } else {
         vm = vmDriverList;
         i = 0;
         outArgs->callerIndex = -1;
         while (vm != NULL && i < vmCount) {
	    if (vm == curVM) {
	       outArgs->callerIndex = i;
	    }
            HostIF_VMLock(vm, 10);
            outArgs->memInfo[i] = vm->memInfo;
            HostIF_VMUnlock(vm, 10);
            i++;
	    vm = vm->nextDriver;
         }
      }
   }

   HostIF_GlobalUnlock(7);

   return TRUE;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Vmx86_GetMemInfoCopy --
 *
 *    Return the information about all VMs by copying the data out to user
 *    memory.
 *
 *    On input, outArgs->numVMs indicates how much space has been allocated for
 *    the information. On output, it indicates how much space has been
 *    filled --hpreg
 *
 * Results:
 *    TRUE on success
 *    FALSE on failure
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

Bool
Vmx86_GetMemInfoCopy(VMDriver *curVM,           // IN
                     VMMemInfoArgs *outArgs)    // IN/OUT
{
   VMMemInfoArgs *buf;
   VMDriver *vm;
   Bool ret = FALSE;

   buf = HostIF_AllocKernelMem(VM_GET_MEM_INFO_SIZE(MAX_VMS), TRUE);
   if (!buf) {
      goto err0;
   }
   if (HostIF_CopyFromUser(buf, outArgs, VM_GET_MEM_INFO_SIZE(1))) {
      goto err1;
   }

   HostIF_GlobalLock(8);

   /* Now that we have the lock, we can read vmCount --hpreg */
   if (buf->numVMs < vmCount) {
      HostIF_GlobalUnlock(8);
      goto err1;
   }

   buf->numLockedPages = numLockedPages;
   buf->lockedPageLimit = lockedPageLimit;
   buf->maxLockedPages = Vmx86LockedPageLimit(curVM);
   buf->globalMinAllocation = Vmx86CalculateGlobalMinAllocation(minVmMemPct);
   buf->minVmMemPct = minVmMemPct;
   buf->callerIndex = -1;
   Vmx86COWStats(&buf->cowInfo);

   for (vm = vmDriverList, buf->numVMs = 0;
        vm;
        vm = vm->nextDriver, buf->numVMs++) {
      ASSERT(buf->numVMs < vmCount);
      if (vm == curVM) {
         buf->callerIndex = buf->numVMs;
      }
      HostIF_VMLock(vm, 11);
      buf->memInfo[buf->numVMs] = vm->memInfo;
      HostIF_VMUnlock(vm, 11);
   }
   ASSERT(buf->numVMs == vmCount);

   HostIF_GlobalUnlock(8);

   if (!HostIF_CopyToUser(outArgs, buf, VM_GET_MEM_INFO_SIZE(buf->numVMs))) {
      ret = TRUE;
   }

err1:
   HostIF_FreeKernelMem(buf);
err0:
   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86SetMemoryUsage --
 *
 *      Updates the paged and nonpaged memory reserved memory values for 
 *      the vm.  
 *
 * Results:
 *      Returns global information about the memory state in args as well
 *      as a value indicating whether or not the virtual machine was
 *      started.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86SetMemoryUsage(VMDriver *curVM, unsigned paged, unsigned nonpaged)
{
   ASSERT(HostIF_VMLockIsHeld(curVM));
   curVM->memInfo.paged         = paged;
   curVM->memInfo.nonpaged      = nonpaged;
   curVM->memInfo.minAllocation = Vmx86MinAllocation(curVM, minVmMemPct);
   curVM->memInfo.maxAllocation = paged + nonpaged;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_Admit --
 *
 *      Set the memory management information about this VM and handles
 *      admission control. We allow vm to power on if there is room for
 *      the minimum allocation for all running vms in memory.  Note that 
 *      the hard memory limit can change dynamically in windows so we
 *      don't have guarantees due to admission control.  
 *
 * Results:
 *      Returns global information about the memory state in args as well
 *      as a value indicating whether or not the virtual machine was
 *      started.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_Admit(VMDriver *curVM,     // IN
            VMMemInfoArgs *args) // IN/OUT
{
   unsigned int globalMinAllocation;

   HostIF_GlobalLock(9);
   /* Update the overcommitment level and minimums for all vms. */
   minVmMemPct = args->minVmMemPct;
   Vmx86UpdateMinAllocations(minVmMemPct);
   globalMinAllocation = Vmx86CalculateGlobalMinAllocation(args->minVmMemPct);

   HostIF_VMLock(curVM, 12);

   curVM->memInfo.shares = args->memInfo->shares;
   curVM->memInfo.usedPct = 100;
   /* mainMemSize is not initialized in WS5.0.  We got size on CREATE_VM, so... */
   if (curVM->vmVersion >= VME_V55) {
      curVM->memInfo.mainMemSize = args->memInfo->mainMemSize;
   }
   Vmx86SetMemoryUsage(curVM, args->memInfo->paged, args->memInfo->nonpaged);
   // Preliminary admission control to put memory pressure on other VMs.
   if (globalMinAllocation + curVM->memInfo.minAllocation <= 
       Vmx86LockedPageLimitForAdmissonControl()) {
      curVM->memInfo.admitted = TRUE;
   } else {
      curVM->memInfo.admitted = FALSE;
   }

#if defined _WIN32
   {
   unsigned int allocatedPages, nonpaged;
   signed int pages;
   MPN32* mpns;
   /* 
    * More admission control: Get enough memory for the nonpaged portion 
    * of the VM.  Drop locks for this long operation.
    * XXX Timeout?
    */
   HostIF_VMUnlock(curVM, 12);
   HostIF_GlobalUnlock(9);

#define ALLOCATE_CHUNK_SIZE 64
   allocatedPages = 0;
   nonpaged = args->memInfo->nonpaged;
   mpns = HostIF_AllocKernelMem(nonpaged * sizeof(MPN32), FALSE);
   if (mpns == NULL) {
      goto undoAdmission;
   }
   while(allocatedPages < nonpaged) {
      pages = Vmx86_AllocLockedPages(curVM, mpns + allocatedPages, 
	                             MIN(ALLOCATE_CHUNK_SIZE, nonpaged - allocatedPages),
				     TRUE);
      if (pages <= 0) {
	 break;
      }
      allocatedPages += pages;
   }

   /* 
    * Free the allocated pages. 
    * XXX Do not free the pages but hand them directly to the admitted VM.
    */

   for (pages = 0; pages < allocatedPages; pages += ALLOCATE_CHUNK_SIZE) {
      Vmx86_FreeLockedPages(curVM, mpns + pages, 
	                    MIN(ALLOCATE_CHUNK_SIZE, allocatedPages - pages), TRUE);
   }
   HostIF_FreeKernelMem(mpns);
#undef ALLOCATE_CHUNK_SIZE

undoAdmission:
   if (allocatedPages != nonpaged) {
       curVM->memInfo.admitted = FALSE; // undo admission
   }

   HostIF_GlobalLock(9);
   HostIF_VMLock(curVM, 12);
   }
#endif

   /* Return global state to the caller. */
   args->memInfo[0] = curVM->memInfo;
   args->numVMs = vmCount;
   args->numLockedPages = numLockedPages;
   args->maxLockedPages = Vmx86LockedPageLimit(curVM);
   args->lockedPageLimit = lockedPageLimit; 
   args->globalMinAllocation = globalMinAllocation;
   args->minVmMemPct = minVmMemPct;
   HostIF_VMUnlock(curVM, 12);
   HostIF_GlobalUnlock(9);
}


Bool
Vmx86_Readmit(VMDriver *curVM, int32 pageDelta)
{
   unsigned globalMinAllocation, newMinAllocation;
   Bool retval = FALSE;
   int32 nonpaged;

   HostIF_GlobalLock(31);
   globalMinAllocation = Vmx86CalculateGlobalMinAllocation(minVmMemPct);
   HostIF_VMLock(curVM, 31);
   nonpaged = curVM->memInfo.nonpaged + pageDelta;
   if (nonpaged >= 0) {
      globalMinAllocation -= Vmx86MinAllocation(curVM, minVmMemPct);
      newMinAllocation = Vmx86MinAllocationFunc(curVM->memInfo.paged,
                                                nonpaged,
                                                curVM->memInfo.mainMemSize,
                                                minVmMemPct);
      if (globalMinAllocation + newMinAllocation <= Vmx86LockedPageLimit(curVM) ||
         pageDelta <= 0) {
         Vmx86SetMemoryUsage(curVM, curVM->memInfo.paged, nonpaged);     
         retval = TRUE;
      }
   }
   HostIF_VMUnlock(curVM, 31);
   HostIF_GlobalUnlock(31);
   return retval;
}
                    
/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetMemUsage --
 *
 *      Set the memory usage by this vm based on its memSample data.
 *
 * Results:
 *      Updates memory allocations.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetMemUsage(VMDriver *curVM, 
		  unsigned usedPct)
{
   if (usedPct <= 100) {
      HostIF_VMLock(curVM, 13);
      curVM->memInfo.usedPct = usedPct;
      HostIF_VMUnlock(curVM, 13);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_PAEEnabled --
 *
 *      Is PAE enabled?
 *
 * Results:
 *      TRUE if PAE enabled.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_PAEEnabled()
{
#ifdef VM_X86_64
   uint64 cr4;
#else
   uint32 cr4;
#endif

   GET_CR4(cr4);
   return (cr4 & CR4_PAE) != 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SVMEnabledCPU --
 * 
 *      Verify that SVM is enabled on the CPU.
 *
 * Results:
 *      TRUE if SVM is enabled, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_SVMEnabledCPU(void)
{
   return SVM_CapableCPU() && SVM_EnabledCPU();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_VTEnabledCPU --
 * 
 *      Verify that VT is enabled on the CPU.
 *
 * Results:
 *      TRUE if VT is enabled, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_VTEnabledCPU(void)
{
   return VT_CapableCPU() && VT_EnabledCPU();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_VTSupportedCPU --
 * 
 *   Verify that the CPU has the VT capabilities required to run the
 *   VT-enabled monitor.
 *
 * Results:
 *      TRUE if the CPU is capable of running the VT-enabled monitor, 
 *      FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
Bool
Vmx86_VTSupportedCPU(Bool use64Bit)
{
   return VT_CapableCPU() && VT_SupportedCPU(use64Bit);
}

/*
 *----------------------------------------------------------------------
 * Vmx86_BrokenCPUHelper --
 *
 *   Process IOCTL_VMX86_BROKEN_CPU_HELPER to Test for a broken CPU.
 *----------------------------------------------------------------------
 */
Bool
Vmx86_BrokenCPUHelper(void)
{
   CPUIDRegs regs;
   uint32 family, model, step;
   uint64 msr_val8b, msr_val17;
   uint32 microcode_version;

   __GET_CPUID(0, &regs);
   /*
    * Currently only have problems with Genuine Intel CPUs:-)
    */
   if ((regs.ebx != 0x756e6547) ||  // Genu
       (regs.ecx != 0x6c65746e) ||  // ntel
       (regs.edx != 0x49656e69)) {  // ineI
      return FALSE;
   }
   __GET_CPUID(1, &regs);
   family = CPUID_FAMILY(regs.eax);
   model = CPUID_MODEL(regs.eax);
   step = CPUID_STEPPING(regs.eax);
   if ((family != 6) || !((model == 7) || (model == 8))) { 
      /* Only have problems with P6 family PIII CPUs*/
      /* Also, we don't know if the below MSR sequence will 
       * work on future Intel processors so we only deal with
       * model 0x67x and 0x68x. 
       */
      return FALSE;
   }

   __SET_MSR(0x8b, 0); // Clear MSR 0x8b
   __GET_CPUID(1, &regs); // Issue CPUID command 1
   msr_val8b = __GET_MSR(0x8b);
   Log("INTELRDMSR(0x8b) = 0x%08x%08x\n",(uint32)(msr_val8b>>32),(uint32)msr_val8b);
   msr_val17 = __GET_MSR(0x17);
   Log("INTELRDMSR(0x17) = 0x%08x%08x\n",(uint32)(msr_val17>>32),(uint32)msr_val17);

   if (!((family == 6) && (model == 8) && (step == 1))) { 
      /* Only have problems with Coppermine CPU cores. */
      return FALSE;
   }

   microcode_version = (uint32)((msr_val8b>>32) & 0x0ff);

   if (microcode_version == 0) { 
      return TRUE;
   }

   return FALSE; // it's good
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_InCompatMode --
 *
 *      See if kernel is running in compatibility mode.
 *
 * Returns:
 *      FALSE if running in full 64-bit mode.
 *      FALSE if running in legacy 32-bit mode.
 *      TRUE if running in compatibility mode.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#ifdef __APPLE__
Bool
Vmx86_InCompatMode(void)
{
   /*
    * This routine only runs on Macs at present, which never run in full 64-bit 
    * mode.  The code below assumes that fact, so we assert it here.
    */
   ASSERT_ON_COMPILE(!vm_x86_64);

   return Vmx86_InLongMode();
}
#endif


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_InLongMode --
 *
 *      See if kernel is running in long (64-bit or compatibility) mode.
 *
 * Returns:
 *      FALSE if running in legacy 32-bit mode.
 *      TRUE if running in full 64-bit mode.
 *      TRUE if running in compatibility mode.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#ifdef __APPLE__
Bool
Vmx86_InLongMode(void)
{
   uint64 efer;

   efer = __GET_MSR(MSR_EFER);
   return (efer & MSR_EFER_LME) != 0;
}
#endif


/*
 *----------------------------------------------------------------------
 *
 * Vmx86LookupVMByID --
 *
 *    Finds the vm with the specified ID.
 *    
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86LookupVMByID(int id)
{
   VMDriver *vm;
   ASSERT(HostIF_GlobalLockIsHeld());
   for (vm = vmDriverList; vm != NULL; vm = vm->nextDriver) {
      if (vm->id == id) {
         return vm;
      }
   }
   return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_FastMonSwitchTransfer --
 *
 *    Gets the old VM from which to transfer pages from, and performs
 *    the transfer.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_FastMonSwitchTransfer(VMDriver *newVM,
                            int       oldVmId)
{
   VMDriver *oldVM = NULL;
   Bool retval = FALSE;
   HostIF_GlobalLock(35);
   oldVM = Vmx86LookupVMByID(oldVmId);
   if (oldVM) {
      if (oldVM->fastMonSwitch.performingSwitch) {
         newVM->fastMonSwitch.oldVmId = oldVmId;
         retval = TRUE;
      } else {
         Warning("Attempted to switch from VM that isn't switching.\n");
      }
   } else {
      Warning("VM lookup failed.\n");
   }
   HostIF_GlobalUnlock(35);
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_FastMonSwitchEnd --
 *
 *    Gets the oldVM and unsets the performingSwitch flag to notify
 *    it that we are done with the fast monitor switch.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_FastMonSwitchEnd(VMDriver *newVM)
{
   VMDriver *oldVM = NULL;
   HostIF_GlobalLock(35);
   oldVM = Vmx86LookupVMByID(newVM->fastMonSwitch.oldVmId);
   if (oldVM) {
      oldVM->fastMonSwitch.performingSwitch = FALSE;
   }
   HostIF_GlobalUnlock(35);
   return oldVM != NULL;
}


#include "vmx86_compat.h"
